"""Storage registry."""
from typing import Type
from mode.utils.imports import FactoryMapping
from faust.types import StoreT

__all__ = ['by_name', 'by_url']

STORES: FactoryMapping[Type[StoreT]] = FactoryMapping(
    memory='faust.stores.memory:Store',
    rocksdb='faust.stores.rocksdb:Store',
)
STORES.include_setuptools_namespace('faust.stores')
by_name = STORES.by_name
by_url = STORES.by_url



#!/usr/bin/env python
import asyncio
import faust

WORDS = ['the', 'quick', 'brown', 'fox']


app = faust.App(
    'word-counts',
    broker='kafka://localhost:9092',
    store='rocksdb://',
    version=1,
    topic_partitions=8,
)

posts_topic = app.topic('posts', value_type=str)
word_counts = app.Table('word_counts', default=int,
                        help='Keep count of words (str to int).')


@app.agent(posts_topic)
async def shuffle_words(posts):
    async for post in posts:
        for word in post.split():
            await count_words.send(key=word, value=word)


@app.agent(value_type=str)
async def count_words(words):
    """Count words from blog post article body."""
    async for word in words:
        word_counts[word] += 1


@app.page('/count/{word}/')
@app.table_route(table=word_counts, match_info='word')
async def get_count(web, request, word):
    return web.json({
        word: word_counts[word],
    })


@app.task
async def sender():
    await posts_topic.maybe_declare()

    for word in WORDS:
        for _ in range(1000):
            await shuffle_words.send(value=word)

    await asyncio.sleep(5.0)
    print(word_counts.as_ansitable(
        key='word',
        value='count',
        title='$$ TALLY $$',
        sort=True,
    ))


@app.on_rebalance_complete.connect
async def on_rebalance_complete(sender, **kwargs):
    print(word_counts.as_ansitable(
        key='word',
        value='count',
        title='$$ TALLY - after rebalance $$',
        sort=True,
    ))


if __name__ == '__main__':
    app.main()



import typing
from docutils import nodes
try:
    from sphinx.errors import NoUri
except ImportError:
    from sphinx.environment import NoUri  # noqa

APPATTRS = {
    'Stream': 'faust.Stream',
    'TableManager': 'faust.tables.TableManager',
    'Serializers': 'faust.serializers.Registry',
    'sensors': 'faust.sensors.SensorDelegate',
    'serializers': 'faust.serializers.Registry',
    'sources': 'faust.topics.TopicManager',
    'tables': 'faust.tables.TableManager',
    'monitor': 'faust.sensors.Monitor',
    'consumer': 'faust.transport.base.Consumer',
    'transport': 'faust.transport.base.Transport',
    'producer': 'faust.transport.base.Producer',

}

APPDIRECT = {
    'client_only',
    'agents',
    'main',
    'topic',
    'agent',
    'task',
    'timer',
    'crontab',
    'stream',
    'Table',
    'Set',
    'start_client',
    'maybe_start_client',
    'send',
    'maybe_start_producer',
    'discover',
    'service',
    'page',
    'command',
}

APPATTRS.update({x: 'faust.App.{0}'.format(x) for x in APPDIRECT})

ABBRS = {
    'App': 'faust.App',
}

ABBR_EMPTY = {
    'exc': 'faust.exceptions',
}
DEFAULT_EMPTY = 'faust.App'


def typeify(S, type):
    if type in ('meth', 'func'):
        return S + '()'
    return S


def shorten(S, newtarget, src_dict):
    if S.startswith('@-'):
        return S[2:]
    elif S.startswith('@'):
        if src_dict is APPATTRS:
            return '.'.join(['app', S[1:]])
        return S[1:]
    return S


def get_abbr(pre, rest, type, orig=None):
    if pre:
        for d in APPATTRS, ABBRS:
            try:
                return d[pre], rest, d
            except KeyError:
                pass
        raise KeyError('Unknown abbreviation: {0} ({1})'.format(
            '.'.join([pre, rest]) if orig is None else orig, type,
        ))
    else:
        for d in APPATTRS, ABBRS:
            try:
                return d[rest], '', d
            except KeyError:
                pass
    return ABBR_EMPTY.get(type, DEFAULT_EMPTY), rest, ABBR_EMPTY


def resolve(S, type):
    if '.' not in S:
        try:
            getattr(typing, S)
        except AttributeError:
            pass
        else:
            return 'typing.{0}'.format(S), None
    orig = S
    if S.startswith('@'):
        S = S.lstrip('@-')
        try:
            pre, rest = S.split('.', 1)
        except ValueError:
            pre, rest = '', S

        target, rest, src = get_abbr(pre, rest, type, orig)
        return '.'.join([target, rest]) if rest else target, src
    return S, None


def pkg_of(module_fqdn):
    return module_fqdn.split('.', 1)[0]


def basename(module_fqdn):
    return module_fqdn.lstrip('@').rsplit('.', -1)[-1]


def modify_textnode(T, newtarget, node, src_dict, type):
    src = node.children[0].rawsource
    return nodes.Text(
        (typeify(basename(T), type) if '~' in src
         else typeify(shorten(T, newtarget, src_dict), type)),
        src,
    )


def maybe_resolve_abbreviations(app, env, node, contnode):
    domainname = node.get('refdomain')
    target = node['reftarget']
    typ = node['reftype']
    if target.startswith('@'):
        newtarget, src_dict = resolve(target, typ)
        node['reftarget'] = newtarget
        # shorten text if '~' is not enabled.
        if len(contnode) and isinstance(contnode[0], nodes.Text):
            contnode[0] = modify_textnode(target, newtarget, node,
                                          src_dict, typ)
        if domainname:
            try:
                domain = env.domains[node.get('refdomain')]
            except KeyError:
                raise NoUri
            return domain.resolve_xref(env, node['refdoc'], app.builder,
                                       typ, newtarget,
                                       node, contnode)


def setup(app):
    app.connect(
        'missing-reference',
        maybe_resolve_abbreviations,
    )
    app.add_crossref_type(
        directivename='sig',
        rolename='sig',
        indextemplate='pair: %s; sig',
    )



import asyncio
import collections.abc
import pytest
from faust.agents import Agent
from faust.agents.actor import Actor, AsyncIterableActor, AwaitableActor
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock


class FakeActor(Actor):
    # need to implement traceback method to test this.

    def traceback(self) -> str:
        return ''


class test_Actor:

    ActorType = FakeActor

    @pytest.fixture()
    def agent(self):
        agent = Mock(name='agent', autospec=Agent)
        agent.name = 'myagent'
        return agent

    @pytest.fixture()
    def stream(self):
        stream = Mock(name='stream')
        stream.active_partitions = None
        return stream

    @pytest.fixture()
    def it(self):
        it = Mock(name='it', autospec=collections.abc.Iterator)
        it.__aiter__ = Mock(name='it.__aiter__')
        it.__await__ = Mock(name='it.__await__')
        return it

    @pytest.fixture()
    def actor(self, *, agent, stream, it):
        return self.ActorType(agent, stream, it)

    def test_constructor(self, *, actor, agent, stream, it):
        assert actor.agent is agent
        assert actor.stream is stream
        assert actor.it is it
        assert actor.index is None
        assert actor.active_partitions is None
        assert actor.actor_task is None

    @pytest.mark.asyncio
    async def test_on_start(self, *, actor):
        actor.actor_task = Mock(name='actor_task', autospec=asyncio.Task)
        actor.add_future = Mock(name='add_future')
        await actor.on_start()
        actor.add_future.assert_called_once_with(actor.actor_task)

    @pytest.mark.asyncio
    async def test_on_stop(self, *, actor):
        actor.cancel = Mock(name='cancel')
        await actor.on_stop()
        actor.cancel.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_isolated_partition_revoked(self, *, actor):
        actor.cancel = Mock(name='cancel')
        actor.stop = AsyncMock(name='stop')
        await actor.on_isolated_partition_revoked(TP('foo', 0))
        actor.cancel.assert_called_once_with()
        actor.stop.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_isolated_partition_assigned(self, *, actor):
        await actor.on_isolated_partition_assigned(TP('foo', 0))

    def test_cancel(self, *, actor):
        actor.actor_task = Mock(name='actor_task', autospec=asyncio.Task)
        actor.cancel()
        actor.stream.channel._throw.assert_called_once()
        actor.stream.channel._throw.call_args[0][0] == StopAsyncIteration()

    def test_cancel__when_no_task(self, *, actor):
        actor.actor_task = None
        actor.cancel()

    def test_repr(self, *, actor):
        assert repr(actor)


class test_AsyncIterableActor(test_Actor):

    ActorType = AsyncIterableActor

    def test_aiter(self, *, actor, it):
        res = actor.__aiter__()
        it.__aiter__.assert_called_with()
        assert res is it.__aiter__()


class test_AwaitableActor(test_Actor):

    ActorType = AwaitableActor

    def test_await(self, *, actor, it):
        res = actor.__await__()
        it.__await__.assert_called_with()
        assert res is it.__await__()



def test_json(faust_json):
    exitcode, agents, stderr = faust_json('agents', '--local')
    assert not exitcode

    assert {'name': '@app.mul',
            'topic': 't-integration-app.mul',
            'help': 'Foo agent help.'} in agents
    assert {'name': '@app.add',
            'topic': 'add-topic',
            'help': '<N/A>'} in agents
    assert {'name': '@app.internal',
            'topic': '<LOCAL>',
            'help': '<N/A>'} in agents

    names = [agent['name'] for agent in agents]
    assert names.index('@app.add') < \
        names.index('@app.internal') < \
        names.index('@app.mul')


def test_tabulated(faust):
    exitcode, stdout, stderr = faust('agents', '--local')
    assert not exitcode
    assert b'@app.mul' in stdout


def test_colors(faust_color):
    exitcode, stdout, stderr = faust_color('agents', '--local')
    assert not exitcode
    assert b'@app.mul' in stdout


def test_json_no_local(faust_json):
    exitcode, agents, stderr = faust_json('agents')
    assert not exitcode

    names = [agent['name'] for agent in agents]
    assert '@app.mul' in names
    assert '@app.add' in names
    assert '@app.internal' not in names



"""Model descriptions.

The model describes the components of a data structure, kind of like a struct
in C, but there's no limitation of what type of data structure the model is,
or what it's used for.

A record (faust.models.record) is a model type that serialize into
dictionaries, so the model describe the fields, and their types:

.. sourcecode:: pycon

    >>> class Point(Record):
    ...    x: int
    ...    y: int

    >>> p = Point(10, 3)
    >>> assert p.x == 10
    >>> assert p.y == 3
    >>> p
    <Point: x=10, y=3>
    >>> payload = p.dumps(serializer='json')
    '{"x": 10, "y": 3, "__faust": {"ns": "__main__.Point"}}'
    >>> p2 = Record.loads(payload)
    >>> p2
    <Point: x=10, y=3>

Models are mainly used for describing the data in messages: both keys and
values can be described as models.
"""
import abc
import warnings

from datetime import datetime
from functools import partial
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
)

from mode.utils.objects import canoname

from faust.exceptions import ValidationError
from faust.serializers.codecs import CodecArg, dumps, loads
from faust.types.models import (
    CoercionMapping,
    FieldDescriptorT,
    FieldMap,
    ModelOptions,
    ModelT,
)

__all__ = ['Model', 'maybe_model', 'registry']

# NOTES:
# - Records are described in the same notation as named tuples in Python 3.6.
#   To accomplish this ``__init_subclass__`` defined in :pep:`487` is used.
#
#   When accessed on the Record class, the attributes are actually field
#   descriptors that return information about the field:
#       >>> Point.x
#       <FieldDescriptor: Point.x: int>
#
#   This field descriptor holds information about the name of the field, the
#   value type of the field, and also what Record subclass it belongs to.
#
# - Sometimes field descriptions are passed around as arguments to functions.
#
# - A stream of deposits may be joined with a stream of orders if
#   both have an ``account`` field.  Field descriptors are used to
#   specify the field.
#
# - order_instance.account is data
#   (it holds the string account for this particular order).
#
# - order_instance.__class__.account is the field descriptor for the field,
#   it's not data but metadata that enables introspection, and it can be
#   passed around to describe a field we want to extract or similar.
#
# - FieldDescriptors are Python descriptors: In Python object
#   attributes can override what happens when they are get/set/deleted:
#
#       class MyDescriptor:
#
#           def __get__(self, instance, cls):
#               if instance is None:
#                   print('ACCESS ON CLASS ATTRIBUTE')
#                   return self
#               print('ACCESS ON INSTANCE')
#               return 42
#
#       class Example:
#           foo = MyDescriptor()
#
#   The above descriptor overrides __get__, which is called when the attribute
#   is accessed (a descriptor may also override __set__ and __del__).
#
#
#   You can see the difference in what happens when you access the attribute
#   on the class, vs. the instance:

#       >>> Example.foo
#       ACCESS ON CLASS ATTRIBUTE
#       <__main__.MyDescriptor at 0x1049caac8>
#
#       >>> x = Example()
#       >>> x.foo
#       ACCESS ON INSTANCE
#       42

E_ABSTRACT_INSTANCE = '''
Cannot instantiate abstract model.

If this model is used as the field of another model,
and you meant to define a polymorphic relationship: make sure
your abstract model class has the `polymorphic_fields` option enabled:

    class {name}(faust.Record, abstract=True, polymorphic_fields=True):
        ...
'''

#: Global map of namespace -> Model, used to find model classes by name.
#: Every single model defined is added here automatically when a model
#: class is defined.
registry: MutableMapping[str, Type[ModelT]] = {}


def maybe_model(arg: Any) -> Any:
    """Convert argument to model if possible."""
    try:
        model = registry[arg['__faust']['ns']]
    except (KeyError, TypeError):
        return arg
    else:
        return model.from_data(arg)


class Model(ModelT):
    """Meta description model for serialization."""

    #: Set to True if this is an abstract base class.
    __is_abstract__: ClassVar[bool] = True

    __validation_errors__ = None

    _pending_finalizers: ClassVar[Optional[List[Callable]]] = None

    #: Serialized data may contain a "blessed key" that mandates
    #: how the data should be deserialized.  This probably only
    #: applies to records, but we need to support it at Model level.
    #: The blessed key has a dictionary value with a ``ns`` key:
    #:   data = {.., '__faust': {'ns': 'examples.simple.Withdrawal'}}
    #: When ``Model._maybe_reconstruct` sees this key it will look
    #: up that namespace in the :data:`registry`, and if it exists
    #: select it as the target model to use for serialization.
    #:
    #: Is this similar to how unsafe deserialization in pickle/yaml/etc.
    #: works?  No! pickle/pyyaml allow for arbitrary types to be
    #: deserialized (and worse in pickle's case), whereas the blessed
    #: key can only deserialize to a hardcoded list of types that are
    #: already under the remote control of messages anyway.
    #: For example it's not possible to perform remote code execution
    #: by providing a blessed key namespace of "os.system", simply
    #: because os.system is not in the registry of allowed types.
    _blessed_key = '__faust'

    @classmethod
    def _maybe_namespace(
            cls, data: Any,
            *,
            preferred_type: Type[ModelT] = None,
            fast_types: Tuple[Type, ...] = (bytes, str),
            isinstance: Callable = isinstance) -> Optional[Type[ModelT]]:
        # The serialized data may contain a ``__faust`` blessed key
        # holding the name of the model it should be deserialized as.
        # So even if value_type=MyModel, the data may mandata that it
        # should be deserialized using "foo.bar.baz" instead.

        # This is how we deal with Kafka's lack of message headers,
        # as needed by the RPC mechanism, without wrapping all data.
        if data is None or isinstance(data, fast_types):
            return None
        try:
            ns = data[cls._blessed_key]['ns']
        except (KeyError, TypeError):
            pass
        else:
            # we only allow blessed keys when type=None, or type=Model
            type_is_abstract = (preferred_type is None or
                                preferred_type is ModelT or
                                preferred_type is Model)
            try:
                model = registry[ns]
            except KeyError:
                if type_is_abstract:
                    raise
                return None
            else:
                if (type_is_abstract or
                        model._options.allow_blessed_key or
                        model._options.polymorphic_fields):
                    return model
        return None

    @classmethod
    def _maybe_reconstruct(cls, data: Any) -> Any:
        model = cls._maybe_namespace(data)
        return model.from_data(data) if model else data

    @classmethod
    def _from_data_field(cls, data: Any) -> Optional['Model']:
        if data is not None:
            if cls.__is_abstract__:
                return cls._maybe_reconstruct(data)
            return cls.from_data(data, preferred_type=cls)
        return None

    @classmethod
    def loads(cls, s: bytes, *,
              default_serializer: CodecArg = None,  # XXX use serializer
              serializer: CodecArg = None) -> ModelT:
        """Deserialize model object from bytes.

        Keyword Arguments:
            serializer (CodecArg): Default serializer to use
                if no custom serializer was set for this model subclass.
        """
        if default_serializer is not None:
            warnings.warn(DeprecationWarning(
                'default_serializer deprecated, use: serializer'))
        ser = cls._options.serializer or serializer or default_serializer
        data = loads(ser, s)
        return cls.from_data(data)

    def __init_subclass__(self,
                          serializer: str = None,
                          namespace: str = None,
                          include_metadata: bool = None,
                          isodates: bool = None,
                          abstract: bool = False,
                          allow_blessed_key: bool = None,
                          decimals: bool = None,
                          coerce: bool = None,
                          coercions: CoercionMapping = None,
                          polymorphic_fields: bool = None,
                          validation: bool = None,
                          date_parser: Callable[[Any], datetime] = None,
                          lazy_creation: bool = False,
                          **kwargs: Any) -> None:
        # Python 3.6 added the new __init_subclass__ function that
        # makes it possible to initialize subclasses without using
        # metaclasses (:pep:`487`).
        super().__init_subclass__(**kwargs)

        # mypy does not recognize `__init_subclass__` as a classmethod
        # and thinks we're mutating a ClassVar when setting:
        #   cls.__is_abstract__ = False
        # To fix this we simply delegate to a _init_subclass classmethod.
        finalizer = partial(
            self._init_subclass,
            serializer,
            namespace,
            include_metadata,
            isodates,
            abstract,
            allow_blessed_key,
            decimals,
            coerce,
            coercions,
            polymorphic_fields,
            validation,
            date_parser,
        )
        if lazy_creation:
            self._pending_finalizers = [finalizer]
        else:
            self._pending_finalizers = None
            finalizer()

    @classmethod
    def make_final(cls) -> None:
        pending, cls._pending_finalizers = cls._pending_finalizers, None
        if pending:
            for finalizer in pending:
                finalizer()

    @classmethod
    def _init_subclass(cls,
                       serializer: str = None,
                       namespace: str = None,
                       include_metadata: bool = None,
                       isodates: bool = None,
                       abstract: bool = False,
                       allow_blessed_key: bool = None,
                       decimals: bool = None,
                       coerce: bool = None,
                       coercions: CoercionMapping = None,
                       polymorphic_fields: bool = None,
                       validation: bool = None,
                       date_parser: Callable[[Any], datetime] = None) -> None:
        # Can set serializer/namespace/etc. using:
        #    class X(Record, serializer='json', namespace='com.vandelay.X'):
        #        ...
        try:
            custom_options = cls.Options
        except AttributeError:
            custom_options = None
        else:
            delattr(cls, 'Options')
        options = getattr(cls, '_options', None)
        if options is None:
            options = ModelOptions()
            options.coercions = {}
            options.defaults = {}
        else:
            options = options.clone_defaults()
        if custom_options:
            options.__dict__.update(custom_options.__dict__)
        if coerce is not None:
            options.coerce = coerce
        if coercions is not None:
            options.coercions.update(coercions)
        if serializer is not None:
            options.serializer = serializer
        if include_metadata is not None:
            options.include_metadata = include_metadata
        if isodates is not None:
            options.isodates = isodates
        if decimals is not None:
            options.decimals = decimals
        if allow_blessed_key is not None:
            options.allow_blessed_key = allow_blessed_key
        if polymorphic_fields is not None:
            options.polymorphic_fields = polymorphic_fields
        if validation is not None:
            options.validation = validation
            options.coerce = True  # validation implies coerce
        if date_parser is not None:
            options.date_parser = date_parser

        options.namespace = namespace or canoname(cls)

        if abstract:
            # Custom base classes can set this to skip class initialization.
            cls.__is_abstract__ = True
            cls._options = options
            cls.__init__ = cls.__abstract_init__  # type: ignore
            return
        cls.__is_abstract__ = False

        # Add introspection capabilities
        cls._contribute_to_options(options)
        # Add FieldDescriptors for every field.
        options.descriptors = cls._contribute_field_descriptors(
            cls, options)

        # Store options on new subclass.
        cls._options = options

        cls._contribute_methods()

        # Register in the global registry, so we can look up
        # models by namespace.
        registry[options.namespace] = cls

        codegens = [
            ('__init__', cls._BUILD_init, '_model_init'),
            ('__hash__', cls._BUILD_hash, '_model_hash'),
            ('__eq__', cls._BUILD_eq, '_model_eq'),
            ('__ne__', cls._BUILD_ne, '_model_ne'),
            ('__gt__', cls._BUILD_gt, '_model_gt'),
            ('__ge__', cls._BUILD_ge, '_model_ge'),
            ('__lt__', cls._BUILD_lt, '_model_lt'),
            ('__le__', cls._BUILD_le, '_model_le'),
        ]

        for meth_name, meth_gen, attr_name in codegens:
            # self._model_init = cls._BUILD_init()
            # if '__init__' not in cls.__dict__:
            #     cls.__init__ = self._model_init
            meth = meth_gen()
            setattr(cls, attr_name, meth)
            if meth_name not in cls.__dict__:
                setattr(cls, meth_name, meth)

    def __abstract_init__(self) -> None:
        raise NotImplementedError(E_ABSTRACT_INSTANCE.format(
            name=type(self).__name__,
        ))

    @classmethod
    @abc.abstractmethod
    def _contribute_to_options(
            cls, options: ModelOptions) -> None:  # pragma: no cover
        ...

    @classmethod
    def _contribute_methods(cls) -> None:  # pragma: no cover
        ...

    @classmethod
    @abc.abstractmethod
    def _contribute_field_descriptors(
            cls,
            target: Type,
            options: ModelOptions,
            parent: FieldDescriptorT = None) -> FieldMap:  # pragma: no cover
        ...

    @classmethod
    @abc.abstractmethod
    def _BUILD_init(cls) -> Callable[[], None]:  # pragma: no cover
        ...

    @classmethod
    @abc.abstractmethod
    def _BUILD_hash(cls) -> Callable[[], None]:  # pragma: no cover
        ...

    @classmethod
    @abc.abstractmethod
    def _BUILD_eq(cls) -> Callable[[], None]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def to_representation(self) -> Any:  # pragma: no cover
        """Convert object to JSON serializable object."""

    @abc.abstractmethod
    def _humanize(self) -> str:  # pragma: no cover
        """Return string representation of object for debugging purposes."""
        ...

    def is_valid(self) -> bool:
        return True if not self.validate() else False

    def validate(self) -> List[ValidationError]:
        errors = self.__validation_errors__
        if errors is None:
            errors = self.__validation_errors__ = list(self._itervalidate())
        return errors

    def validate_or_raise(self) -> None:
        errors = self.validate()
        if errors:
            raise errors[0]

    def _itervalidate(self) -> Iterable[ValidationError]:
        for name, descr in self._options.descriptors.items():
            yield from descr.validate_all(getattr(self, name))

    @property
    def validation_errors(self) -> List[ValidationError]:
        return self.validate()

    def derive(self, *objects: ModelT, **fields: Any) -> ModelT:
        """Derive new model with certain fields changed."""
        return self._derive(*objects, **fields)

    @abc.abstractmethod  # pragma: no cover
    def _derive(self, *objects: ModelT, **fields: Any) -> ModelT:
        raise NotImplementedError()

    def dumps(self, *, serializer: CodecArg = None) -> bytes:
        """Serialize object to the target serialization format."""
        return dumps(serializer or self._options.serializer,
                     self.to_representation())

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self._humanize()}>'



import pytest
from mode.utils.mocks import Mock


@pytest.fixture()
def context(*, app, request):
    context = Mock()
    context.app = app
    context.find_root.return_value = context
    return context



import datetime
import faust
import pytest

from faust.events import Event
from faust.types import Message
from faust.tables.wrappers import WindowSet, WindowWrapper
from mode.utils.mocks import Mock, patch


class TableKey(faust.Record):
    value: str


class TableValue(faust.Record):
    id: str
    num: int


KEY1 = TableKey('foo')
VALUE1 = TableValue('id', 3)

WINDOW1 = faust.HoppingWindow(size=10, step=2, expires=3600.0)


def event():
    message = Message(topic='test-topic', key='key', value='value',
                      partition=3, offset=0, checksum=None,
                      timestamp=datetime.datetime.now().timestamp(),
                      timestamp_type=0,
                      headers={})
    return Event(
        app='test-app',
        key='key',
        value='value',
        headers={},
        message=message,
    )


class test_Table:

    @pytest.fixture
    def table(self, *, app):
        return self.create_table(app, name='foo', default=int)

    @pytest.fixture
    def strict_table(self, *, app):
        return self.create_table(app, name='strict')

    def create_table(self, app, *,
                     name='foo',
                     key_type=TableKey,
                     value_type=TableValue,
                     **kwargs):
        return app.Table(
            name,
            key_type=key_type,
            value_type=value_type,
            **kwargs)

    @patch('faust.tables.wrappers.current_event', return_value=event())
    def test_using_window(self, patch_current, *, table):
        with_wrapper = table.using_window(WINDOW1)
        self.assert_wrapper(with_wrapper, table, WINDOW1)
        self.assert_current(with_wrapper, patch_current)

    @patch('faust.tables.wrappers.current_event', return_value=event())
    def test_hopping(self, patch_current, *, table):
        with_wrapper = table.hopping(10, 2, 3600)
        self.assert_wrapper(with_wrapper, table)
        self.assert_current(with_wrapper, patch_current)

    @patch('faust.tables.wrappers.current_event', return_value=event())
    def test_tumbling(self, patch_current, *, table):
        with_wrapper = table.tumbling(10, 3600)
        self.assert_wrapper(with_wrapper, table)
        self.assert_current(with_wrapper, patch_current)

    def assert_wrapper(self, wrapper, table, window=None):
        assert wrapper.table is table
        t = wrapper.table
        if window is not None:
            assert t.window is window
        assert t._changelog_compacting
        assert t._changelog_deleting
        assert t._changelog_topic is None
        assert isinstance(wrapper, WindowWrapper)

    def assert_current(self, wrapper, patch_current):
        value = wrapper['test']
        assert isinstance(value, WindowSet)
        patch_current.asssert_called_once_with()
        assert value.current() == 0

    def test_missing__when_default(self, *, table):
        assert table['foo'] == 0
        table.data['foo'] = 3
        assert table['foo'] == 3

    def test_missing__no_default(self, *, strict_table):
        with pytest.raises(KeyError):
            strict_table['foo']
        strict_table.data['foo'] = 3
        assert strict_table['foo'] == 3

    def test_has_key(self, *, table):
        assert not table._has_key('foo')
        table.data['foo'] = 3
        assert table._has_key('foo')

    def test_get_key(self, *, table):
        assert table._get_key('foo') == 0
        table.data['foo'] = 3
        assert table._get_key('foo') == 3

    def test_set_key(self, *, table):
        with patch('faust.tables.base.current_event') as current_event:
            event = current_event.return_value
            partition = event.message.partition
            table.send_changelog = Mock(name='send_changelog')
            table._set_key('foo', 'val')
            table.send_changelog.asssert_called_once_with(
                partition, 'foo', 'val')
            assert table['foo'] == 'val'

    def test_del_key(self, *, table):
        with patch('faust.tables.base.current_event') as current_event:
            event = current_event.return_value
            partition = event.message.partition
            table.send_changelog = Mock(name='send_changelog')
            table.data['foo'] = 3
            table._del_key('foo')
            table.send_changelog.asssert_called_once_with(
                partition, 'foo', None)
            assert 'foo' not in table.data

    def test_as_ansitable(self, *, table):
        table.data['foo'] = 'bar'
        table.data['bar'] = 'baz'
        assert table.as_ansitable(sort=True)
        assert table.as_ansitable(sort=False)

    def test_on_key_set__no_event(self, *, table):
        with patch('faust.tables.base.current_event') as ce:
            ce.return_value = None
            with pytest.raises(TypeError):
                table.on_key_set('k', 'v')

    def test_on_key_del__no_event(self, *, table):
        with patch('faust.tables.base.current_event') as ce:
            ce.return_value = None
            with pytest.raises(TypeError):
                table.on_key_del('k')



from decimal import Decimal
import pytest
from mode.utils.mocks import Mock
from faust import Record
from faust.exceptions import ValidationError
from faust.models.fields import (
    BooleanField,
    BytesField,
    DecimalField,
    FieldDescriptor,
)


class X(Record):
    foo: str


class test_ValidationError:

    @pytest.fixture()
    def field(self):
        return DecimalField(model=X, field='foo')

    @pytest.fixture()
    def error(self, *, field):
        return ValidationError('error', field=field)

    def test_repr(self, *, error):
        assert repr(error)

    def test_str(self, *, error):
        assert str(error)


class test_FieldDescriptor:

    def test_validate(self):
        f = FieldDescriptor()
        assert list(f.validate('foo')) == []


class test_BooleanField:

    @pytest.fixture()
    def model(self):
        model = Mock(name='model')
        model.__name__ = 'Model'
        return model

    @pytest.fixture()
    def field(self, *, model):
        return self._new_field(model, required=True)

    def _new_field(self, model, required: bool, **kwargs) -> BooleanField:
        return BooleanField(
            field='foo',
            type=bool,
            required=True,
            model=model,
            coerce=True,
            **kwargs,
        )

    @pytest.mark.parametrize('value', [
        True,
        False,
    ])
    def test_validate_bool(self, value, *, field):
        assert not list(field.validate(value))

    @pytest.mark.parametrize('value', [
        '',
        None,
        12,
        3.2,
        object,
    ])
    def test_validate_other(self, value, *, field):
        errors = list(field.validate(value))
        assert errors
        assert str(errors[0]).startswith(
            'foo must be True or False, of type bool')

    @pytest.mark.parametrize('value,expected', [
        ('', False),
        ('foo', True),
        (0, False),
        (1, True),
        (999, True),
        (object(), True),
        (None, False),
        ({}, False),
        ([], False),
        ([1], True),
    ])
    def test_prepare_value__when_coerce(self, value, expected, *, field):
        assert field.prepare_value(value) is expected

    def test_prepare_value__no_coerce(self, *, field):
        assert field.prepare_value(None, coerce=False) is None


class test_DecimalField:

    def test_init_options(self):
        assert DecimalField(max_digits=3).max_digits == 3
        assert DecimalField(max_decimal_places=4).max_decimal_places == 4

        f = DecimalField(max_digits=3, max_decimal_places=4)
        f2 = f.clone()
        assert f2.max_digits == 3
        assert f2.max_decimal_places == 4

        f3 = DecimalField()
        assert f3.max_digits is None
        assert f3.max_decimal_places is None
        f4 = f3.clone()
        assert f4.max_digits is None
        assert f4.max_decimal_places is None

    @pytest.mark.parametrize('value', [
        Decimal('Inf'),
        Decimal('NaN'),
        Decimal('sNaN'),
    ])
    def test_infinite(self, value):
        f = DecimalField(coerce=True, field='foo')
        with pytest.raises(ValidationError):
            raise next(f.validate(value))

    @pytest.mark.parametrize('value,places,digits', [
        (Decimal(4.1), 100, 2),
        (Decimal(4.1), 100, 2),
        (Decimal(4.1), None, 2),
        (Decimal(4.12), 100, None),
        (Decimal(4.123), 100, None),
        (4.1234, 100, 2),
        (Decimal(4.1234), 100, 2),
        (Decimal(123456612341.1234), 100, 100),
    ])
    def test_max_decimal_places__good(self, value, places, digits):
        f = DecimalField(
            max_decimal_places=places,
            max_digits=digits,
            coerce=True,
            field='foo',
        )
        d: Decimal = f.prepare_value(value)
        for error in f.validate(d):
            raise error

    @pytest.mark.parametrize('value', [
        Decimal(1.12412421421),
        Decimal(1.12345),
        Decimal(123456788.12345),
    ])
    def test_max_decimal_places__bad(self, value):
        f = DecimalField(max_decimal_places=4, coerce=True, field='foo')
        with pytest.raises(ValidationError):
            raise next(f.validate(value))

    @pytest.mark.parametrize('value', [
        Decimal(12345.12412421421),
        Decimal(123456.12345),
        Decimal(123456788.12345),
    ])
    def test_max_digits__bad(self, value):
        f = DecimalField(max_digits=4, coerce=True, field='foo')
        with pytest.raises(ValidationError):
            raise next(f.validate(value))


class test_BytesField:

    def test_init_options(self):
        assert BytesField(encoding='latin1').encoding == 'latin1'
        assert BytesField(errors='replace').errors == 'replace'

        f = BytesField(encoding='latin1', errors='replace')
        f2 = f.clone()
        assert f2.encoding == 'latin1'
        assert f2.errors == 'replace'

        f3 = BytesField()
        assert f3.encoding == 'utf-8'
        assert f3.errors == 'strict'
        f4 = f3.clone()
        assert f4.encoding == 'utf-8'
        assert f4.errors == 'strict'

    @pytest.mark.parametrize('value,coerce,trim,expected_result', [
        ('foo', True, False, b'foo'),
        (b'foo', True, False, b'foo'),
        ('foo', False, False, 'foo'),
        ('  fo o   ', True, True, b'fo o'),
        (b'  fo o   ', True, True, b'fo o'),
        ('  fo o   ', True, False, b'  fo o   '),
    ])
    def test_prepare_value(self, value, coerce, trim, expected_result):
        f = BytesField(coerce=coerce, trim_whitespace=trim)
        assert f.prepare_value(value) == expected_result



import pytest
from faust import Event
from faust.stores.memory import Store
from faust.types import TP
from mode.utils.mocks import Mock


class test_Store:

    @pytest.fixture
    def store(self, *, app):
        return Store(url='memory://', app=app, table=Mock(name='table'))

    def test_clear(self, *, store):
        store.data['foo'] = 1
        store._clear()
        assert not store.data

    def test_apply_changelog_batch(self, *, store):
        event, to_key, to_value = self.mock_event_to_key_value()
        store.apply_changelog_batch([event], to_key=to_key, to_value=to_value)

        to_key.assert_called_once_with(b'key')
        to_value.assert_called_once_with(b'value')
        assert store.data[to_key()] == to_value()

    def test_apply_changelog_batch__deletes_key_for_None_value(self, *, store):
        self.test_apply_changelog_batch(store=store)
        event2, to_key, to_value = self.mock_event_to_key_value(value=None)

        assert to_key() in store.data
        store.apply_changelog_batch([event2], to_key=to_key, to_value=to_value)

        assert to_key() not in store.data

    def mock_event_to_key_value(self, key=b'key', value=b'value'):
        event = self.mock_event(key=key, value=value)
        to_key, to_value = self.mock_to_key_value(event)
        return event, to_key, to_value

    def mock_event(self, key=b'key', value=b'value'):
        event = Mock(name='event', autospec=Event)
        event.key = key
        event.value = value
        event.message.key = key
        event.message.value = value
        return event

    def mock_to_key_value(self, event):
        to_key = Mock(name='to_key')
        to_key.return_value = event.key
        to_value = Mock(name='to_value')
        to_value.return_value = event.value
        return to_key, to_value

    def test_persisted_offset(self, *, store):
        assert store.persisted_offset(TP('foo', 0)) is None

    def test_reset_state(self, *, store):
        store.reset_state()



"""Parsing ISO-8601 string and converting to :class:`~datetime.datetime`."""
try:  # pragma: no cover
    import ciso8601
except ImportError:  # pragma: no cover
    from ._iso8601_python import parse
else:
    parse = ciso8601.parse_datetime

__all__ = ['parse']



#!/usr/bin/env python
"""Withdrawal example.

Quick Start
===========

1) Start worker:

.. sourcecode:: console

    $ ./examples/simple.py worker -l info

2) Start sending example data:

    $ ./examples/simple.py produce
"""
import asyncio
import random
from datetime import datetime, timezone
from itertools import count
import faust
from faust.cli import option


class Withdrawal(faust.Record, isodates=True, serializer='json'):
    user: str
    country: str
    amount: float
    date: datetime = None


app = faust.App(
    'faust-withdrawals29',
    broker='kafka://127.0.0.1:9092',
    store='rocksdb://',
    origin='examples.withdrawals',
    topic_partitions=4,
)
withdrawals_topic = app.topic('withdrawals129', value_type=Withdrawal)
foo_topic = app.topic('withdrawalz129', value_type=Withdrawal)

withdrawal_counts = app.Table('withdrawal-counts', default=int)
seen_events = 0


@app.agent(withdrawals_topic, isolated_partitions=True)
async def track_user_withdrawal(withdrawals):
    global seen_events
    i = 0
    async for _withdrawal in withdrawals:  # noqa
        event = withdrawals.current_event
        assert event.message.tp in withdrawals.active_partitions
        i += 1
        seen_events += 1
        withdrawal_counts[_withdrawal.user] += 1
        # Uncomment to raise exceptions during processing
        # to test agents restarting.
        #  if not i % 3:
        #      raise KeyError('OH NO')


@app.task
async def report_progress(app):
    prev_count = 0
    while not app.should_stop:
        await app.sleep(5.0)
        if seen_events <= prev_count:
            print(f'Not progressing: was {prev_count} now {seen_events}')
        else:
            print(f'Progressing: was {prev_count} now {seen_events}')
        prev_count = seen_events


@app.command(
    option('--max-latency',
           type=float, default=0.5, envvar='PRODUCE_LATENCY',
           help='Add delay of (at most) n seconds between publishing.'),
    option('--max-messages',
           type=int, default=None,
           help='Send at most N messages or 0 for infinity.'),
)
async def produce(self, max_latency: float, max_messages: int):
    """Produce example Withdrawal events."""
    for i, withdrawal in enumerate(generate_withdrawals(max_messages)):
        await withdrawals_topic.send(key=withdrawal.user, value=withdrawal)
        if not i % 10000:
            self.say(f'+SEND {i}')
        if max_latency:
            await asyncio.sleep(random.uniform(0, max_latency))


def generate_withdrawals(n: int = None):
    for d in generate_withdrawals_dict(n):
        yield Withdrawal(**d)


def generate_withdrawals_dict(n: int = None):
    num_countries = 5
    countries = [f'country_{i}' for i in range(num_countries)]
    country_dist = [0.9] + ([0.10 / num_countries] * (num_countries - 1))
    num_users = 500
    users = [f'user_{i}' for i in range(num_users)]
    for _ in range(n) if n is not None else count():
        yield {
            'user': random.choice(users),
            'amount': random.uniform(0, 25_000),
            'country': random.choices(countries, country_dist)[0],
            'date': datetime.utcnow().replace(tzinfo=timezone.utc).isoformat(),
        }


if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        sys.argv.extend(['worker', '-l', 'info'])
    app.main()



import enum
from collections import Counter
from datetime import date, datetime, timezone
from decimal import Decimal
from uuid import uuid4
from faust.utils.json import JSONEncoder, str_to_decimal
from hypothesis import assume, given
from hypothesis.strategies import decimals
import pytest


@given(decimals())
def test_str_to_decimal_decimals(x):
    assume(x.is_finite())
    assert str_to_decimal(str(x)) == x


def test_str_to_decimal_None():
    assert str_to_decimal(None) is None


def test_str():
    d1 = Decimal('3.3333433434343434343434343434343')
    assert str_to_decimal(str(d1)) == d1


def test_maxlen():
    s = '3.' + '34' * 1000
    with pytest.raises(ValueError):
        str_to_decimal(s)


def test_NaN():
    with pytest.raises(ValueError):
        str_to_decimal('NaN')


def test_Inf():
    with pytest.raises(ValueError):
        str_to_decimal('Inf')


def test_negative_Inf():
    with pytest.raises(ValueError):
        str_to_decimal('-Inf')


def test_sNaN():
    with pytest.raises(ValueError):
        str_to_decimal('sNaN')


class Flags(enum.Enum):
    X = 'Xval'
    Y = 'Yval'


class CanJson:

    def __json__(self):
        return 'yes'


def test_JSONEncoder():
    encoder = JSONEncoder()
    assert encoder.default(date(2016, 3, 2))
    assert encoder.default(datetime.utcnow())
    assert encoder.default(datetime.now(timezone.utc))
    assert encoder.default(Counter([('foo', 3), ('bar', 4)]))
    assert encoder.default(uuid4())
    assert encoder.default(Flags.X) == 'Xval'
    assert encoder.default(Flags.Y) == 'Yval'
    assert encoder.default({1, 2, 3}) == [1, 2, 3]
    assert encoder.default(CanJson()) == 'yes'
    with pytest.raises(TypeError):
        encoder.default(object())



from typing import Mapping
from mode import setup_logging
import pytest
from .app import simple


setup_logging(loglevel='INFO')


def _build_data(i: int) -> Mapping:
    return {'A': {'the': {'quick': {'brown': {'fox': i}}}}}


@pytest.mark.asyncio
async def test_simple_ask() -> None:
    for i in range(100):
        value = _build_data(i)
        assert await simple.ask(value=value) == value


@pytest.mark.asyncio
async def test_simple_map() -> None:
    values = [_build_data(i) for i in range(100)]
    check = set(range(100))
    replies = set()
    async for reply in simple.map(values):
        v = reply['A']['the']['quick']['brown']['fox']
        assert v in check
        replies.add(v)
    assert replies == check



import pytest
from faust.sensors.statsd import StatsdMonitor

PREFIX = 'faust-test'


@pytest.fixture()
def mon():
    return StatsdMonitor(prefix=PREFIX)


def stream_from_channel(app):
    return app.stream(app.channel())


def stream_from_topic(app):
    return app.stream(app.topic('withdrawals'))


def stream_from_multiple_topics(app):
    return app.stream(app.topic('foo', 'bar'))


def stream_from_combined_streams(app):
    return app.stream(app.topic('foo')) & app.stream(app.topic('bar'))


@pytest.mark.parametrize('make_stream,expected', [
    (stream_from_channel, 'channel_anon'),
    (stream_from_topic, 'topic_withdrawals'),
    (stream_from_multiple_topics, 'topic_foo,bar'),
    (stream_from_combined_streams, 'topic_foo&topic_bar'),
])
def test_stream_label(make_stream, expected, *, app, mon):
    assert mon._stream_label(make_stream(app)) == expected



import asyncio
from mode.utils.aiter import anext
import pytest
from t.helpers import message

__all__ = ['channel_empty', 'times_out', 'is_empty', 'message', 'put']


async def channel_empty(channel, *, timeout=0.01):
    assert channel.empty()
    with pytest.raises(asyncio.TimeoutError):
        await channel.get(timeout=timeout)
    return True


async def times_out(coro, *, timeout=0.01):
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(coro, timeout=timeout)
    return True


async def is_empty(it, *, timeout=0.01):
    return await times_out(anext(it), timeout=timeout)


async def put(channel, key=None, value=None, **kwargs):
    msg = message(key=key, value=value, **kwargs)
    await channel.deliver(msg)
    return msg



#!/usr/bin/env python

# In this exapmple we have a function `publish_every_2secs` publishing a
# message every 2 seconds to topic `hopping_topic`.
# We have created an agent `print_windowed_events` consuming events from
# `hopping_topic` that mutates the windowed table `hopping_table`.

# `hopping_table` is a table with hopping (overlaping) windows. Each of
# its windows is 10 seconds of duration, and we create a new window every 5
# seconds.
# |----------|
#       |-----------|
#             |-----------|
#                   |-----------|
# Since we produce an event every 2 seconds and our windows are 10
# seconds of duration we expect different the following results per method
# called in `WindowWrapper`:
# - now(): Gets the closest value to current local time. It will always be
# between 1 and 3.
# - current(): Gets the value relative to the event's timestamp. It will
# always be between 1 and 3.
# - value(): Gets the value relative to default relative option. It will
# always be between 1 and 3.
# - delta(30): Gets the value of window 30 secs before the current event. For
# the first 20 seconds it will be 0 and after second 30 it will always be 5.

from random import random
from datetime import timedelta
import faust

app = faust.App('windowing', broker='kafka://localhost:9092')


class Model(faust.Record, serializer='json'):
    random: float


TOPIC = 'hopping_topic'

hopping_topic = app.topic(TOPIC, value_type=Model)
hopping_table = app.Table(
    'hopping_table',
    default=int).hopping(10, 5, expires=timedelta(minutes=10))


@app.agent(hopping_topic)
async def print_windowed_events(stream):
    async for _ in stream: # noqa
        hopping_table['counter'] += 1
        value = hopping_table['counter']
        print(f'-- New Event (every 2 secs) written to hopping(10, 5) --')
        print(f'now() should have values between 1-3: {value.now()}')
        print(f'current() should have values between 1-3: {value.current()}')
        print(f'value() should have values between 1-3: {value.value()}')
        print(f'delta(30) should start at 0 and after 40 secs be 5: '
              f'{value.delta(30)}')


@app.timer(2.0, on_leader=True)
async def publish_every_2secs():
    msg = Model(random=round(random(), 2))
    await hopping_topic.send(value=msg)


if __name__ == '__main__':
    app.main()



"""Base-interface for sensors."""
from time import monotonic
from typing import Any, Dict, Iterator, Mapping, Optional, Set

from mode import Service

from faust import web
from faust.types import AppT, CollectionT, EventT, StreamT
from faust.types.assignor import PartitionAssignorT
from faust.types.tuples import Message, PendingMessage, RecordMetadata, TP
from faust.types.sensors import SensorDelegateT, SensorT
from faust.types.topics import TopicT
from faust.types.transports import ConsumerT, ProducerT

__all__ = ['Sensor', 'SensorDelegate']


class Sensor(SensorT, Service):
    """Base class for sensors.

    This sensor does not do anything at all, but can be subclassed
    to create new monitors.
    """

    def on_message_in(self, tp: TP, offset: int, message: Message) -> None:
        """Message received by a consumer."""
        ...

    def on_stream_event_in(self, tp: TP, offset: int, stream: StreamT,
                           event: EventT) -> Optional[Dict]:
        """Message sent to a stream as an event."""
        return None

    def on_stream_event_out(self, tp: TP, offset: int, stream: StreamT,
                            event: EventT, state: Dict = None) -> None:
        """Event was acknowledged by stream.

        Notes:
            Acknowledged means a stream finished processing the event, but
            given that multiple streams may be handling the same event,
            the message cannot be committed before all streams have
            processed it.  When all streams have acknowledged the event,
            it will go through :meth:`on_message_out` just before offsets
            are committed.
        """
        ...

    def on_message_out(self,
                       tp: TP,
                       offset: int,
                       message: Message) -> None:
        """All streams finished processing message."""
        ...

    def on_topic_buffer_full(self, topic: TopicT) -> None:
        """Topic buffer full so conductor had to wait."""
        ...

    def on_table_get(self, table: CollectionT, key: Any) -> None:
        """Key retrieved from table."""
        ...

    def on_table_set(self, table: CollectionT, key: Any, value: Any) -> None:
        """Value set for key in table."""
        ...

    def on_table_del(self, table: CollectionT, key: Any) -> None:
        """Key deleted from table."""
        ...

    def on_commit_initiated(self, consumer: ConsumerT) -> Any:
        """Consumer is about to commit topic offset."""
        ...

    def on_commit_completed(self, consumer: ConsumerT, state: Any) -> None:
        """Consumer finished committing topic offset."""
        ...

    def on_send_initiated(self, producer: ProducerT, topic: str,
                          message: PendingMessage,
                          keysize: int, valsize: int) -> Any:
        """About to send a message."""
        ...

    def on_send_completed(self,
                          producer: ProducerT,
                          state: Any,
                          metadata: RecordMetadata) -> None:
        """Message successfully sent."""
        ...

    def on_send_error(self,
                      producer: ProducerT,
                      exc: BaseException,
                      state: Any) -> None:
        """Error while sending message."""
        ...

    def on_assignment_start(self,
                            assignor: PartitionAssignorT) -> Dict:
        """Partition assignor is starting to assign partitions."""
        return {'time_start': monotonic()}

    def on_assignment_error(self,
                            assignor: PartitionAssignorT,
                            state: Dict,
                            exc: BaseException) -> None:
        """Partition assignor did not complete assignor due to error."""
        ...

    def on_assignment_completed(self,
                                assignor: PartitionAssignorT,
                                state: Dict) -> None:
        """Partition assignor completed assignment."""
        ...

    def on_rebalance_start(self, app: AppT) -> Dict:
        """Cluster rebalance in progress."""
        return {'time_start': monotonic()}

    def on_rebalance_return(self, app: AppT, state: Dict) -> None:
        """Consumer replied assignment is done to broker."""
        ...

    def on_rebalance_end(self, app: AppT, state: Dict) -> None:
        """Cluster rebalance fully completed (including recovery)."""
        ...

    def on_web_request_start(self, app: AppT, request: web.Request, *,
                             view: web.View = None) -> Dict:
        """Web server started working on request."""
        return {'time_start': monotonic()}

    def on_web_request_end(self,
                           app: AppT,
                           request: web.Request,
                           response: Optional[web.Response],
                           state: Dict,
                           *,
                           view: web.View = None) -> None:
        """Web server finished working on request."""
        ...

    def asdict(self) -> Mapping:
        """Convert sensor state to dictionary."""
        return {}


class SensorDelegate(SensorDelegateT):
    """A class that delegates sensor methods to a list of sensors."""

    _sensors: Set[SensorT]

    def __init__(self, app: AppT) -> None:
        self.app = app
        self._sensors = set()

    def add(self, sensor: SensorT) -> None:
        """Add sensor."""
        # connect beacons
        sensor.beacon = self.app.beacon.new(sensor)
        self._sensors.add(sensor)

    def remove(self, sensor: SensorT) -> None:
        """Remove sensor."""
        self._sensors.remove(sensor)

    def __iter__(self) -> Iterator:
        return iter(self._sensors)

    def on_message_in(self, tp: TP, offset: int, message: Message) -> None:
        """Call before message is delegated to streams."""
        for sensor in self._sensors:
            sensor.on_message_in(tp, offset, message)

    def on_stream_event_in(self, tp: TP, offset: int, stream: StreamT,
                           event: EventT) -> Optional[Dict]:
        """Call when stream starts processing an event."""
        return {
            sensor: sensor.on_stream_event_in(tp, offset, stream, event)
            for sensor in self._sensors
        }

    def on_stream_event_out(self, tp: TP, offset: int, stream: StreamT,
                            event: EventT, state: Dict = None) -> None:
        """Call when stream is done processing an event."""
        sensor_state = state or {}
        for sensor in self._sensors:
            sensor.on_stream_event_out(tp, offset, stream, event,
                                       sensor_state.get(sensor))

    def on_topic_buffer_full(self, topic: TopicT) -> None:
        """Call when conductor topic buffer is full and has to wait."""
        for sensor in self._sensors:
            sensor.on_topic_buffer_full(topic)

    def on_message_out(self,
                       tp: TP,
                       offset: int,
                       message: Message) -> None:
        """Call when message is fully acknowledged and can be committed."""
        for sensor in self._sensors:
            sensor.on_message_out(tp, offset, message)

    def on_table_get(self, table: CollectionT, key: Any) -> None:
        """Call when value in table is retrieved."""
        for sensor in self._sensors:
            sensor.on_table_get(table, key)

    def on_table_set(self, table: CollectionT, key: Any, value: Any) -> None:
        """Call when new value for key in table is set."""
        for sensor in self._sensors:
            sensor.on_table_set(table, key, value)

    def on_table_del(self, table: CollectionT, key: Any) -> None:
        """Call when key in a table is deleted."""
        for sensor in self._sensors:
            sensor.on_table_del(table, key)

    def on_commit_initiated(self, consumer: ConsumerT) -> Any:
        """Call when consumer commit offset operation starts."""
        # This returns arbitrary state, so we return a map from sensor->state.
        return {
            sensor: sensor.on_commit_initiated(consumer)
            for sensor in self._sensors
        }

    def on_commit_completed(self, consumer: ConsumerT, state: Any) -> None:
        """Call when consumer commit offset operation completed."""
        # state is now a mapping from sensor->state, so
        # make sure to correct the correct state to each sensor.
        for sensor in self._sensors:
            sensor.on_commit_completed(consumer, state[sensor])

    def on_send_initiated(self, producer: ProducerT, topic: str,
                          message: PendingMessage,
                          keysize: int, valsize: int) -> Any:
        """Call when message added to producer buffer."""
        return {
            sensor: sensor.on_send_initiated(
                producer, topic, message, keysize, valsize)
            for sensor in self._sensors
        }

    def on_send_completed(self,
                          producer: ProducerT,
                          state: Any,
                          metadata: RecordMetadata) -> None:
        """Call when producer finished sending message."""
        for sensor in self._sensors:
            sensor.on_send_completed(producer, state[sensor], metadata)

    def on_send_error(self,
                      producer: ProducerT,
                      exc: BaseException,
                      state: Any) -> None:
        """Call when producer was unable to publish message."""
        for sensor in self._sensors:
            sensor.on_send_error(producer, exc, state[sensor])

    def on_assignment_start(self,
                            assignor: PartitionAssignorT) -> Dict:
        """Partition assignor is starting to assign partitions."""
        return {
            sensor: sensor.on_assignment_start(assignor)
            for sensor in self._sensors
        }

    def on_assignment_error(self,
                            assignor: PartitionAssignorT,
                            state: Dict,
                            exc: BaseException) -> None:
        """Partition assignor did not complete assignor due to error."""
        for sensor in self._sensors:
            sensor.on_assignment_error(assignor, state[sensor], exc)

    def on_assignment_completed(self,
                                assignor: PartitionAssignorT,
                                state: Dict) -> None:
        """Partition assignor completed assignment."""
        for sensor in self._sensors:
            sensor.on_assignment_completed(assignor, state[sensor])

    def on_rebalance_start(self, app: AppT) -> Dict:
        """Cluster rebalance in progress."""
        return {
            sensor: sensor.on_rebalance_start(app)
            for sensor in self._sensors
        }

    def on_rebalance_return(self, app: AppT, state: Dict) -> None:
        """Consumer replied assignment is done to broker."""
        for sensor in self._sensors:
            sensor.on_rebalance_return(app, state[sensor])

    def on_rebalance_end(self, app: AppT, state: Dict) -> None:
        """Cluster rebalance fully completed (including recovery)."""
        for sensor in self._sensors:
            sensor.on_rebalance_end(app, state[sensor])

    def on_web_request_start(self, app: AppT, request: web.Request, *,
                             view: web.View = None) -> Dict:
        """Web server started working on request."""
        return {
            sensor: sensor.on_web_request_start(app, request,
                                                view=view)
            for sensor in self._sensors
        }

    def on_web_request_end(self,
                           app: AppT,
                           request: web.Request,
                           response: Optional[web.Response],
                           state: Dict,
                           *,
                           view: web.View = None) -> None:
        """Web server finished working on request."""
        for sensor in self._sensors:
            sensor.on_web_request_end(app, request, response, state[sensor],
                                      view=view)

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self._sensors!r}>'



import operator
import random
from datetime import datetime

import faust
import pytest
from faust.events import Event
from faust.exceptions import ImproperlyConfigured
from faust.tables.wrappers import WindowSet
from faust.types import Message
from mode.utils.mocks import Mock, patch

DATETIME = datetime.utcnow()
DATETIME_TS = DATETIME.timestamp()


class User(faust.Record):
    id: str
    name: str


@pytest.fixture
def table(*, app):
    return app.Table('name')


@pytest.fixture
def wtable(*, table):
    return table.hopping(60, 1, 3600.0)


@pytest.fixture
def iwtable(*, table):
    return table.hopping(60, 1, 3600.0, key_index=True)


@pytest.fixture
def event():
    return Mock(name='event', autospec=Event)


def same_items(a, b):
    a_list = _maybe_items(a)
    b_list = _maybe_items(b)
    return same(a_list, b_list)


def _maybe_items(d):
    try:
        items = d.items
    except AttributeError:
        return d
    else:
        return items()


def same(a, b):
    return sorted(a) == sorted(b)


@pytest.yield_fixture()
def current_event(*, freeze_time):
    with patch('faust.tables.wrappers.current_event') as current_event:
        with patch('faust.tables.base.current_event', current_event):
            current_event.return_value.message.timestamp = freeze_time.time
            yield current_event


class test_WindowSet:

    @pytest.fixture
    def wset(self, *, wtable, event):
        return WindowSet('k', wtable.table, wtable, event)

    def test_constructor(self, *, event, table, wset, wtable):
        assert wset.key == 'k'
        assert wset.table is table
        assert wset.wrapper is wtable
        assert wset.event is event

    def test_apply(self, *, wset, event):
        Mock(name='event2', autospec=Event)
        wset.wrapper.get_timestamp = Mock(name='wrapper.get_timestamp')
        wset.table._apply_window_op = Mock(name='_apply_window_op')

        ret = wset.apply(operator.add, 'val')
        wset.wrapper.get_timestamp.assert_called_once_with(wset.event)
        wset.table._apply_window_op.assert_called_once_with(
            operator.add, 'k', 'val', wset.wrapper.get_timestamp(),
        )
        assert ret is wset

    def mock_get_timestamp(self, wset):
        m = wset.wrapper.get_timestamp = Mock(name='wrapper.get_timestamp')
        return m

    def test_apply__custom_event(self, *, wset, event):
        event2 = Mock(name='event2', autospec=Event)
        wset.table._apply_window_op = Mock(name='_apply_window_op')
        get_timestamp = self.mock_get_timestamp(wset)

        ret = wset.apply(operator.add, 'val', event2)
        get_timestamp.assert_called_once_with(event2)
        wset.table._apply_window_op.assert_called_once_with(
            operator.add, 'k', 'val', get_timestamp(),
        )
        assert ret is wset

    def test_value(self, *, event, wset):
        get_timestamp = self.mock_get_timestamp(wset)
        wset.table._windowed_timestamp = Mock(name='_windowed_timestamp')
        assert wset.value(event)
        wset.table._windowed_timestamp.assert_called_once_with(
            'k', get_timestamp())

    def test_now(self, *, wset):
        wset.table._windowed_now = Mock(name='_windowed_now')
        ret = wset.now()
        wset.table._windowed_now.assert_called_once_with('k')
        assert ret is wset.table._windowed_now()

    def test_current(self, *, table, wset):
        event2 = Mock(name='event2', autospec=Event)
        table._windowed_timestamp = Mock(name='_windowed_timestamp')
        table._relative_event = Mock(name='_relative_event')
        ret = wset.current(event2)
        table._relative_event.assert_called_once_with(event2)
        table._windowed_timestamp.assert_called_once_with(
            'k', table._relative_event())
        assert ret is table._windowed_timestamp()

    def test_current__default_event(self, *, table, wset):
        table._windowed_timestamp = Mock(name='_windowed_timestamp')
        table._relative_event = Mock(name='_relative_event')
        ret = wset.current()
        table._relative_event.assert_called_once_with(wset.event)
        table._windowed_timestamp.assert_called_once_with(
            'k', table._relative_event())
        assert ret is table._windowed_timestamp()

    def test_delta(self, *, table, wset):
        event2 = Mock(name='event2', autospec=Event)
        table._windowed_delta = Mock(name='_windowed_delta')
        ret = wset.delta(30.3, event2)
        table._windowed_delta.assert_called_once_with('k', 30.3, event2)
        assert ret is table._windowed_delta()

    def test_delta__default_event(self, *, table, wset):
        table._windowed_delta = Mock(name='_windowed_delta')
        ret = wset.delta(30.3)
        table._windowed_delta.assert_called_once_with('k', 30.3, wset.event)
        assert ret is table._windowed_delta()

    def test_getitem(self, *, wset):
        wset.table = {(wset.key, 30.3): 101.1}
        assert wset[30.3] == 101.1

    def test_getitem__event(self, *, app, wset):
        e = Event(app,
                  key='KK',
                  value='VV',
                  headers={},
                  message=Mock(name='message', autospec=Message))
        ret = wset[e]
        assert isinstance(ret, WindowSet)
        assert ret.key == wset.key
        assert ret.table is wset.table
        assert ret.wrapper is wset.wrapper
        assert ret.event is e

    def test_setitem(self, *, wset):
        wset.table = {}
        wset[30.3] = 'val'
        assert wset.table[(wset.key, 30.3)] == 'val'

    def test_setitem__event(self, *, app, wset):
        e = Event(app,
                  key='KK',
                  value='VV',
                  headers={},
                  message=Mock(name='message', autospec=Message))
        with pytest.raises(NotImplementedError):
            wset[e] = 'val'

    def test_delitem(self, *, wset):
        wset.table = {(wset.key, 30.3): 'val'}
        del(wset[30.3])
        assert not wset.table

    def test_delitem__event(self, *, app, wset):
        e = Event(app,
                  key='KK',
                  value='VV',
                  headers={},
                  message=Mock(name='message', autospec=Message))
        with pytest.raises(NotImplementedError):
            del(wset[e])

    @pytest.mark.parametrize('meth,expected_op', [
        ('__iadd__', operator.add),
        ('__isub__', operator.sub),
        ('__imul__', operator.mul),
        ('__itruediv__', operator.truediv),
        ('__ifloordiv__', operator.floordiv),
        ('__imod__', operator.mod),
        ('__ipow__', operator.pow),
        ('__ilshift__', operator.lshift),
        ('__irshift__', operator.rshift),
        ('__iand__', operator.and_),
        ('__ixor__', operator.xor),
        ('__ior__', operator.or_),
    ])
    def test_operators(self, meth, expected_op, *, wset):
        other = Mock(name='other')
        op = getattr(wset, meth)
        wset.apply = Mock(name='apply')
        result = op(other)

        wset.apply.assert_called_once_with(expected_op, other)
        assert result is wset.apply()

    def test_repr(self, *, wset):
        assert repr(wset)


class test_WindowWrapper:

    def test_name(self, *, wtable):
        assert wtable.name == wtable.table.name

    def test_relative_to(self, *, wtable):
        relative_to = Mock(name='relative_to')
        w2 = wtable.relative_to(relative_to)
        assert w2.table is wtable.table
        assert w2._get_relative_timestamp is relative_to

    def test_relative_to_now(self, *, table, wtable):
        w2 = wtable.relative_to_now()
        assert w2._get_relative_timestamp == wtable.table._relative_now

    def test_relative_to_field(self, *, table, wtable):
        table._relative_field = Mock(name='_relative_field')
        field = Mock(name='field')
        w2 = wtable.relative_to_field(field)
        table._relative_field.assert_called_once_with(field)
        assert w2._get_relative_timestamp == table._relative_field()

    def test_relative_to_stream(self, *, table, wtable):
        w2 = wtable.relative_to_stream()
        assert w2._get_relative_timestamp == wtable.table._relative_event

    @pytest.mark.parametrize('input,expected', [
        (DATETIME, DATETIME_TS),
        (303.333, 303.333),
        (None, 99999.6),
    ])
    def test_get_timestamp(self, input, expected, *, event, wtable):
        event.message.timestamp = 99999.6
        if input is not None:
            wtable.get_relative_timestamp = lambda e=None: input
        else:
            wtable.get_relative_timestamp = None
        assert wtable.get_timestamp(event) == expected

    def test_get_timestamp__event_is_None(self, *, event, wtable):
        wtable.get_relative_timestamp = None
        with patch('faust.tables.wrappers.current_event') as ce:
            ce.return_value = None
            with pytest.raises(RuntimeError):
                assert wtable.get_timestamp(None)

    def test_on_recover(self, *, wtable, table):
        cb = Mock(name='callback')
        wtable.on_recover(cb)
        assert cb in table._recover_callbacks

    def test_contains(self, *, table, wtable):
        table._windowed_contains = Mock(name='windowed_contains')
        wtable.get_timestamp = Mock(name='get_timestamp')
        ret = wtable.__contains__('k')
        wtable.get_timestamp.assert_called_once_with()
        table._windowed_contains.assert_called_once_with(
            'k', wtable.get_timestamp())
        assert ret is table._windowed_contains()

    def test_getitem(self, *, wtable):
        w = wtable['k2']
        assert isinstance(w, WindowSet)
        assert w.key == 'k2'
        assert w.table is wtable.table
        assert w.wrapper is wtable

    def test_setitem(self, *, table, wtable):
        table._set_windowed = Mock(name='set_windowed')
        wtable.get_timestamp = Mock(name='get_timestamp')
        wtable['foo'] = 300
        wtable.get_timestamp.assert_called_once_with()
        table._set_windowed.assert_called_once_with(
            'foo', 300, wtable.get_timestamp(),
        )

    def test_setitem__key_is_WindowSet(self, *, wtable):
        wtable['k2'] = wtable['k2']

    def test_delitem(self, *, table, wtable):
        table._del_windowed = Mock(name='del_windowed')
        wtable.get_timestamp = Mock(name='get_timestamp')
        del(wtable['foo'])
        wtable.get_timestamp.assert_called_once_with()
        table._del_windowed.assert_called_once_with(
            'foo', wtable.get_timestamp(),
        )

    def test_len__no_key_index_raises(self, *, wtable):
        with pytest.raises(NotImplementedError):
            len(wtable)

    def test_as_ansitable__raises(self, *, wtable):
        with pytest.raises(NotImplementedError):
            wtable.as_ansitable()

    def test_keys_raises(self, *, wtable):
        with pytest.raises(NotImplementedError):
            list(wtable._keys())

    @pytest.mark.parametrize('input', [
        datetime.now(),
        103.33,
        User.id,
        lambda s: s,
    ])
    def test_relative_handler(self, input, *, wtable):
        wtable.get_relative_timestamp = input
        assert wtable.get_relative_timestamp

    def test_relative_handler__invalid_handler(self, *, wtable):
        with pytest.raises(ImproperlyConfigured):
            wtable._relative_handler(object())


class test_WindowWrapper_using_key_index:
    TABLE_DATA = {
        'foobar': 'AUNIQSTR',
        'xuzzy': 'BUNIQSTR',
    }

    TABLE_DATA_DELTA = {
        'foobar': 'AUNIQSTRdelta1',
        'xuzzy': 'BUNIQSTRdelta1',
    }

    @pytest.fixture
    def wset(self, *, iwtable, event):
        return WindowSet('k', iwtable.table, iwtable, event)

    @pytest.fixture()
    def data(self, *, freeze_time, iwtable):
        iwtable.key_index_table = {k: 1 for k in self.TABLE_DATA}
        iwtable.table._data = {}
        for w in iwtable.table._window_ranges(freeze_time.time):
            iwtable.table._data.update({
                (k, w): v
                for k, v in self.TABLE_DATA.items()
            })
        return iwtable.table._data

    @pytest.fixture()
    def data_with_30s_delta(self, *, freeze_time, iwtable, data):
        window = iwtable.table.window
        for key, value in self.TABLE_DATA.items():
            data[(key, window.delta(freeze_time.time, 30))] = value + 'delta1'

    @pytest.fixture()
    def remove_a_key(self, *, iwtable, data):
        remove_key = random.choice(list(self.TABLE_DATA))
        items_leftover = {
            k: v for
            k, v in self.TABLE_DATA.items()
            if k != remove_key
        }
        iwtable.table._data = {
            k: v
            for k, v in iwtable.table._data.items()
            if k[0] != remove_key
        }
        return items_leftover

    def test_len(self, *, iwtable):
        iwtable.key_index_table = {1: 'A', 2: 'B'}
        assert len(iwtable) == 2

    def test_as_ansitable(self, *, iwtable, data):
        table = iwtable.relative_to_now().as_ansitable()
        print(table)
        assert table
        assert 'foobar' in table
        assert 'AUNIQSTR' in table

    def test_items(self, *, iwtable, data):
        assert same_items(iwtable.relative_to_now(), self.TABLE_DATA)

    def test_items_keys_in_index_not_in_table(self, *, iwtable, remove_a_key):
        assert same_items(iwtable.relative_to_now(), remove_a_key)

    def test_items_now(self, *, iwtable, data):
        assert same_items(iwtable.items().now(), self.TABLE_DATA)

    def test_items_now_keys_in_index_not_in_table(
            self, *, iwtable, remove_a_key):
        assert same_items(iwtable.items().now(), remove_a_key)

    def test_items_current(self, *, iwtable, data, current_event):
        assert same_items(iwtable.items().current(), self.TABLE_DATA)

    def test_items_current_keys_in_index_not_in_table(
            self, *, iwtable, remove_a_key, current_event):
        assert same_items(iwtable.items().current(), remove_a_key)

    def test_items_delta(self, *, iwtable, data_with_30s_delta, current_event):
        assert same_items(iwtable.items().delta(30), self.TABLE_DATA_DELTA)

    def test_items_delta_key_not_in_table(
            self, *, iwtable,
            data_with_30s_delta, remove_a_key, current_event):
        expected = {
            k: v for k, v in self.TABLE_DATA_DELTA.items()
            if k in remove_a_key
        }
        assert same_items(iwtable.items().delta(30), expected)

    def test_keys(self, *, iwtable, data):
        assert same(iwtable.relative_to_now().keys(), self.TABLE_DATA)

    def test_keys__now(self, *, iwtable, data):
        assert same(iwtable.relative_to_now().keys().now(), self.TABLE_DATA)

    def test_keys__current(self, *, iwtable, data, current_event):
        keys = iwtable.relative_to_now().keys().current()
        assert same(keys, self.TABLE_DATA)

    def test_keys__delta(self, *, iwtable, data, current_event):
        keys = iwtable.relative_to_now().keys().delta(1000)
        assert same(keys, [])
        keys = iwtable.relative_to_now().keys().delta(10)
        assert same(keys, self.TABLE_DATA)

    def test_iter(self, *, iwtable, data):
        assert same(iwtable.relative_to_now(), self.TABLE_DATA)

    def test_values(self, *, iwtable, data):
        assert same(iwtable.relative_to_now().values(),
                    self.TABLE_DATA.values())

    def test_values_keys_in_index_not_in_table(self, *, iwtable, remove_a_key):
        assert same(iwtable.relative_to_now().values(),
                    remove_a_key.values())

    def test_values_now(self, *, iwtable, data):
        assert same(iwtable.values().now(), self.TABLE_DATA.values())

    def test_values_now_keys_in_index_not_in_table(
            self, *, iwtable, remove_a_key):
        assert same(iwtable.values().now(), remove_a_key.values())

    def test_values_current(self, *, iwtable, data, current_event):
        assert same(iwtable.values().current(), self.TABLE_DATA.values())

    def test_values_current_keys_in_index_not_in_table(
            self, *, iwtable, remove_a_key, current_event):
        assert same(iwtable.values().current(), remove_a_key.values())

    def test_values_delta(
            self, *, iwtable, data_with_30s_delta, current_event):
        assert same(iwtable.values().delta(30), self.TABLE_DATA_DELTA.values())

    def test_values_delta_key_not_in_table(
            self, *, iwtable,
            data_with_30s_delta, remove_a_key, current_event):
        expected = {
            k: v for k, v in self.TABLE_DATA_DELTA.items()
            if k in remove_a_key
        }
        assert same(iwtable.values().delta(30), expected.values())

    def test_setitem(self, *, wset):
        wset.table = {}
        wset.wrapper.key_index_table = {}
        wset[30.3] = 'val'
        assert wset.table[(wset.key, 30.3)] == 'val'
        assert wset.key in wset.wrapper.key_index_table
        wset[30.3] = 'val2'
        assert wset.table[(wset.key, 30.3)] == 'val2'

    def test_delitem(self, *, wset):
        wset.table = {(wset.key, 30.3): 'val'}
        wset.wrapper.key_index_table = {wset.key: 1}
        del(wset[30.3])
        assert not wset.table
        assert not wset.wrapper.key_index_table



"""HTTP endpoint showing partition routing destinations."""
from faust import web
from faust.web.exceptions import ServiceUnavailable

__all__ = ['TableList', 'TableDetail', 'TableKeyDetail', 'blueprint']

blueprint = web.Blueprint('router')


@blueprint.route('/', name='list')
class TableList(web.View):
    """List routes for all tables."""

    async def get(self, request: web.Request) -> web.Response:
        """Return JSON response with list of all table routes."""
        router = self.app.router
        return self.json(router.tables_metadata())


@blueprint.route('/{name}/', name='detail')
class TableDetail(web.View):
    """List route for specific table."""

    async def get(self, request: web.Request, name: str) -> web.Response:
        """Return JSON response with table metadata."""
        router = self.app.router
        return self.json(router.table_metadata(name))


@blueprint.route('/{name}/{key}/', name='key-detail')
class TableKeyDetail(web.View):
    """List information about key."""

    async def get(self,
                  request: web.Request,
                  name: str,
                  key: str) -> web.Response:
        """Return JSON response after looking up the route of a table key.

        Arguments:
            name: Name of table.
            key: Key to look up node for.

        Raises:
            faust.web.exceptions.ServiceUnavailable: if broker metadata
                has not yet been received.
        """
        router = self.app.router
        try:
            dest_url = router.key_store(name, key)
        except KeyError:
            raise ServiceUnavailable()
        else:
            return self.json(str(dest_url))



"""Program ``faust send`` used to send events to agents and topics."""
import asyncio
import random
from typing import Any
from faust.types import CodecArg, K, RecordMetadata, V
from .base import AppCommand, argument, option

__all__ = ['send']


class send(AppCommand):
    """Send message to agent/topic."""

    topic: Any
    key: K
    key_serializer: CodecArg
    value: V
    value_serializer: CodecArg
    repeat: int
    min_latency: float
    max_latency: float

    options = [
        option('--key-type', '-K',
               help='Name of model to serialize key into.'),
        option('--key-serializer',
               help='Override default serializer for key.'),
        option('--value-type', '-V',
               help='Name of model to serialize value into.'),
        option('--value-serializer',
               help='Override default serializer for value.'),
        option('--key', '-k',
               help='String value for key (use json if model).'),
        option('--partition', type=int,
               help='Specific partition to send to.'),
        option('--repeat', '-r', type=int, default=1,
               help='Send message n times.'),
        option('--min-latency', type=float, default=0.0,
               help='Minimum delay between sending.'),
        option('--max-latency', type=float, default=0.0,
               help='Maximum delay between sending.'),

        argument('entity'),
        argument('value', default=None, required=False),
    ]

    async def run(self,
                  entity: str,
                  value: str,
                  *args: Any,
                  key: str = None,
                  key_type: str = None,
                  key_serializer: str = None,
                  value_type: str = None,
                  value_serializer: str = None,
                  partition: int = 1,
                  timestamp: float = None,
                  repeat: int = 1,
                  min_latency: float = 0.0,
                  max_latency: float = 0.0,
                  **kwargs: Any) -> Any:
        """Send message to topic/agent/channel."""
        if key is not None:
            key = self.to_key(key_type, key)
        if value is not None:
            value = self.to_value(value_type, value)
        topic = self.to_topic(entity)
        for i in range(repeat):
            self.carp(f'k={key!r} v={value!r} -> {topic!r}...')
            fut_send_complete = await topic.send(
                key=key,
                value=value,
                partition=partition,
                timestamp=timestamp,
                key_serializer=key_serializer,
                value_serializer=value_serializer,
            )
            meta: RecordMetadata = await fut_send_complete
            self.say(self.dumps(meta._asdict()))
            if i and max_latency:
                await asyncio.sleep(random.uniform(min_latency, max_latency))
        await self.app.producer.stop()



import pytest
from faust.web import Blueprint
from faust.web.base import (
    BlueprintManager,
    DEBUG_BLUEPRINTS,
    DEFAULT_BLUEPRINTS,
    PRODUCTION_BLUEPRINTS,
    Web,
)
from mode.utils.mocks import Mock, patch
from yarl import URL


class test_BlueprintManager:

    @pytest.fixture()
    def manager(self):
        return BlueprintManager()

    def test_add(self, *, manager):
        bp1 = Blueprint('test1')
        manager.add('/v1/', bp1)
        assert manager._enabled
        assert manager._enabled[0] == ('/v1/', bp1)

    def test_add__already_applied(self, *, manager):
        bp1 = Blueprint('test1')
        manager.applied = True
        with pytest.raises(RuntimeError):
            manager.add('/v1', bp1)

    def test_apply(self, *, manager):
        web = Mock(name='web')
        bp1 = Mock(name='blueprint1')
        bp1.name = 'blueprint1'
        bp2 = Mock(name='blueprint2')
        bp2.name = 'blueprint2'

        manager.add('/v1/', bp1)
        manager.add('/v2/', bp2)
        manager.apply(web)

        assert manager._active['blueprint1'] is bp1
        assert manager._active['blueprint2'] is bp2

        bp1.register.assert_called_once_with(web.app, url_prefix='/v1/')
        bp2.register.assert_called_once_with(web.app, url_prefix='/v2/')
        bp1.init_webserver.assert_called_once_with(web)
        bp2.init_webserver.assert_called_once_with(web)

        assert manager.applied
        manager.apply(web)


class MyWeb(Web):

    def text(self, *args, **kwargs):
        ...

    def html(self, *args, **kwargs):
        ...

    def json(self, *args, **kwargs):
        ...

    def bytes(self, *args, **kwargs):
        ...

    def bytes_to_response(self, *args, **kwargs):
        ...

    def response_to_bytes(self, *args, **kwargs):
        ...

    def route(self, *args, **kwargs):
        ...

    def add_static(self, *args, **kwargs):
        ...

    async def read_request_content(self, *args, **kwargs):
        ...

    async def wsgi(self, *args, **kwargs):
        ...


class test_Web:

    @pytest.fixture()
    def web(self, *, app):
        return MyWeb(app)

    @pytest.mark.conf(debug=True)
    def test_debug_blueprints(self, *, web):
        assert web.app.conf.debug
        assert web.blueprints._enabled == (
            DEFAULT_BLUEPRINTS + DEBUG_BLUEPRINTS)

    def test_production_blueprints(self, *, web):
        assert not web.app.conf.debug
        assert web.blueprints._enabled == (
            DEFAULT_BLUEPRINTS + PRODUCTION_BLUEPRINTS)

    def test_url_for(self, *, web):
        web.reverse_names['test'] = '/foo/{bar}/'
        assert web.url_for(
            'test', bar='the quick/fox') == '/foo/the%20quick%2Ffox/'

    def test_url_for__not_found(self, *, web):
        with pytest.raises(KeyError):
            web.url_for('foo')

    def test_url__on_localhost(self, *, web, app):
        with patch('socket.gethostname') as gethostname:
            gethostname.return_value = 'foobar.example.com'
            app.conf.web_port = 3030
            app.conf.canonical_url = URL('http://foobar.example.com')
            assert web.url == URL('http://localhost:3030')

    def test_url__not_on_localhost(self, *, web, app):
        with patch('socket.gethostname') as gethostname:
            gethostname.return_value = 'foobar.example.com'
            app.conf.canonical_url = URL('http://xuzzy.example.com')
            assert web.url == URL('http://xuzzy.example.com')



import os
import subprocess
import sys
from pathlib import Path
from typing import Callable, Tuple
from faust.types import AppT
from faust.utils.json import loads
import pytest
from t.integration import app as _app_module


@pytest.fixture
def app() -> AppT:
    os.environ.pop('F_DATADIR', None)
    os.environ.pop('FAUST_DATADIR', None)
    os.environ.pop('F_WORKDIR', None)
    os.environ.pop('FAUST_WORKDIR', None)
    return _app_module.app


@pytest.fixture
def main_path() -> Path:
    return Path(_app_module.__file__).with_suffix('.py')


CommandReturns = Tuple[int, str, str]


def _create_faust_cli(executable: Path, *partial_args: str,
                      color: bool = False,
                      json: bool = False) -> Callable[..., CommandReturns]:
    if not color:
        partial_args += ('--no-color',)
    if json:
        partial_args += ('--json',)

    def call_faust_cli(*args: str) -> Tuple[str, str]:
        p = subprocess.Popen(
            [sys.executable,
             str(executable)] + list(partial_args) + list(args),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
        )
        stdout, stderr = p.communicate()
        if json:
            print(f'JSON RET: {p.returncode} {stdout!r} {stderr!r}')
            ret = p.returncode, loads(stdout), stderr
            return ret
        print(f'TEXT RET: {p.returncode} {stdout!r} {stderr!r}')
        return p.returncode, stdout, stderr
    return call_faust_cli


@pytest.fixture
def faust(main_path: Path) -> Callable[..., CommandReturns]:
    return _create_faust_cli(main_path)


@pytest.fixture
def faust_json(main_path: Path):
    return _create_faust_cli(main_path, json=True)


@pytest.fixture
def faust_color(main_path: Path) -> Callable[..., CommandReturns]:
    return _create_faust_cli(main_path, color=True)



"""Python :pypi:`click` parameter types."""
from typing import Any, Iterable, Optional
import click
from click.types import ParamType, StringParamType
from yarl import URL

__all__ = [
    'WritableDirectory',
    'WritableFilePath',
    'CaseInsensitiveChoice',
    'TCPPort',
    'URLParam',
]


WritableDirectory = click.Path(
    exists=False,      # create if needed,
    file_okay=False,   # must be directory,
    dir_okay=True,     # not file;
    writable=True,     # and read/write access.
    readable=True,     #
)

WritableFilePath = click.Path(
    exists=False,       # create if needed,
    file_okay=True,     # must be file,
    dir_okay=False,     # not directory;
    writable=True,      # and read/write access.
    readable=True,      #
)


class CaseInsensitiveChoice(click.Choice):
    """Case-insensitive version of :class:`click.Choice`."""

    def __init__(self, choices: Iterable[Any]) -> None:
        self.choices = [str(val).lower() for val in choices]
        self.case_sensitive = False  # required by click 7.x

    def convert(self,
                value: str,
                param: Optional[click.Parameter],
                ctx: Optional[click.Context]) -> Any:
        """Convert string to case-insensitive choice."""
        if value.lower() in self.choices:
            return value
        return super().convert(value, param, ctx)


class TCPPort(click.IntRange):
    """CLI option: TCP Port (integer in range 1 - 65535)."""

    name = 'range[1-65535]'

    def __init__(self) -> None:
        super().__init__(1, 65535)


class URLParam(ParamType):
    """URL :pypi:`click` parameter type.

    Converts any string URL to :class:`yarl.URL`.
    """

    name = 'URL'

    _string_param: StringParamType

    def __init__(self) -> None:
        self._string_param = StringParamType()

    def convert(self,
                value: str,
                param: Optional[click.Parameter],
                ctx: Optional[click.Context]) -> URL:
        """Convert :class:`str` argument to :class:`yarl.URL`."""
        text_value = self._string_param.convert(value, param, ctx)
        return URL(text_value)

    def __repr__(self) -> str:
        return 'URL'



"""Registry of supported codecs (serializers, compressors, etc.)."""
import sys

from decimal import Decimal
from typing import Any, Optional, Tuple, Type, cast

from mode.utils.compat import want_bytes, want_str
from mode.utils.objects import cached_property

from faust.exceptions import KeyDecodeError, ValueDecodeError
from faust.types import K, ModelArg, ModelT, V
from faust.types.serializers import RegistryT

from .codecs import CodecArg, dumps, loads

__all__ = ['Registry']

IsInstanceArg = Tuple[Type, ...]


class Registry(RegistryT):
    """Serializing message keys/values.

    Arguments:
        key_serializer: Default key serializer to use when none provided.
        value_serializer: Default value serializer to use when none provided.
    """

    def __init__(self,
                 key_serializer: CodecArg = None,
                 value_serializer: CodecArg = 'json') -> None:
        self.key_serializer = key_serializer
        self.value_serializer = value_serializer

    def loads_key(self,
                  typ: Optional[ModelArg],
                  key: Optional[bytes],
                  *,
                  serializer: CodecArg = None) -> K:
        """Deserialize message key.

        Arguments:
            typ: Model to use for deserialization.
            key: Serialized key.

            serializer: Codec to use for this value.  If not set
               the default will be used (:attr:`key_serializer`).
        """
        if key is None:
            if typ is not None and issubclass(typ, ModelT):
                raise KeyDecodeError(f'Expected {typ!r}, received {key!r}!')
            return key
        serializer = serializer or self.key_serializer
        try:
            payload = self._loads(serializer, key)
            return cast(K, self._prepare_payload(typ, payload))
        except MemoryError:
            raise
        except Exception as exc:
            raise KeyDecodeError(str(exc)).with_traceback(
                sys.exc_info()[2]) from exc

    def _loads(self, serializer: CodecArg, data: bytes) -> Any:
        return loads(serializer, data)

    def _serializer(self, typ: Optional[ModelArg], *alt: CodecArg) -> CodecArg:
        serializer = None
        for serializer in alt:
            if serializer:
                break
        else:
            if typ is str:
                return 'raw'
            elif typ is bytes:
                return 'raw'
        return serializer

    def loads_value(self,
                    typ: Optional[ModelArg],
                    value: Optional[bytes],
                    *,
                    serializer: CodecArg = None) -> Any:
        """Deserialize value.

        Arguments:
            typ: Model to use for deserialization.
            value: bytes to deserialize.

            serializer: Codec to use for this value.  If not set
               the default will be used (:attr:`value_serializer`).
        """
        if value is None:
            if typ is not None and issubclass(typ, ModelT):
                raise ValueDecodeError(
                    f'Expected {typ!r}, received {value!r}!')
            return None
        serializer = self._serializer(typ, serializer, self.value_serializer)
        try:
            payload = self._loads(serializer, value)
            return cast(V, self._prepare_payload(typ, payload))
        except MemoryError:
            raise
        except Exception as exc:
            raise ValueDecodeError(str(exc)).with_traceback(
                sys.exc_info()[2]) from exc

    def _prepare_payload(self, typ: Optional[ModelArg], value: Any) -> Any:
        if typ is None:  # (autodetect)
            return self.Model._maybe_reconstruct(value)
        elif typ is int:
            return int(want_str(value))
        elif typ is float:
            return float(want_str(value))
        elif typ is Decimal:
            return Decimal(want_str(value))
        elif typ is str:
            return want_str(value)
        elif typ is bytes:
            return want_bytes(value)
        else:
            # type set to Model
            model = cast(ModelT, typ)
            return model.from_data(value, preferred_type=model)

    def dumps_key(self,
                  typ: Optional[ModelArg],
                  key: K,
                  *,
                  serializer: CodecArg = None,
                  skip: IsInstanceArg = (bytes,)) -> Optional[bytes]:
        """Serialize key.

        Arguments:
            typ: Model hint (can also be :class:`str`/:class:`bytes`).
                 When ``typ=str`` or :class:`bytes`, raw serializer is assumed.
            key: The key value to serializer.

            serializer: Codec to use for this key, if it is not a model type.
               If not set the default will be used (:attr:`key_serializer`).
        """
        is_model = False
        if isinstance(key, ModelT):
            is_model = True
            serializer = key._options.serializer or serializer
        serializer = self._serializer(typ, serializer, self.key_serializer)
        if serializer and not isinstance(key, skip):
            if is_model:
                return cast(ModelT, key).dumps(serializer=serializer)
            return dumps(serializer, key)
        return want_bytes(cast(bytes, key)) if key is not None else None

    def dumps_value(self,
                    typ: Optional[ModelArg],
                    value: V,
                    *,
                    serializer: CodecArg = None,
                    skip: IsInstanceArg = (bytes,)) -> Optional[bytes]:
        """Serialize value.

        Arguments:
            typ: Model hint (can also be :class:`str`/:class:`bytes`).
                 When ``typ=str`` or :class:`bytes`, raw serializer is assumed.
            key: The value to serializer.

            serializer: Codec to use for this value, if it is not a model type.
              If not set the default will be used (:attr:`value_serializer`).
        """
        is_model = False
        if isinstance(value, ModelT):
            is_model = True
            serializer = value._options.serializer or serializer
        serializer = self._serializer(typ, serializer, self.value_serializer)
        if serializer and not isinstance(value, skip):
            if is_model:
                return cast(ModelT, value).dumps(serializer=serializer)
            return dumps(serializer, value)
        return cast(bytes, value)

    @cached_property
    def Model(self) -> Type[ModelT]:
        """Return the :class:`faust.Model` class used by this serializer."""
        from faust.models.base import Model
        return Model



import asyncio
import pytest
from typing import Hashable
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock


class test_AgentManager:

    def create_agent(self, name, topic_names=None):
        agent = Mock(
            name=name,
            maybe_start=AsyncMock(),
            stop=AsyncMock(),
            restart=AsyncMock(),
            on_partitions_revoked=AsyncMock(),
            on_partitions_assigned=AsyncMock(),
            get_topic_names=Mock(return_value=topic_names),
        )
        return agent

    @pytest.fixture()
    def agents(self, *, app):
        agents = app.agents
        return agents

    @pytest.fixture()
    def agent1(self):
        return self.create_agent('agent1', ['t1'])

    @pytest.fixture()
    def agent2(self):
        return self.create_agent('agent2', ['t1', 't2', 't3'])

    @pytest.fixture()
    def many(self, *, agents, agent1, agent2):
        agents['foo'] = agent1
        agents['bar'] = agent2
        return agents

    def test_constructor(self, *, agents, app):
        assert agents.app is app
        assert agents.data == {}
        assert agents._by_topic == {}

    @pytest.mark.asyncio
    async def test_on_stop__agent_raises_cancel(self, *, many, agent1, agent2):
        agent1.stop.coro.side_effect = asyncio.CancelledError()
        agent2.stop.coro.side_effect = asyncio.CancelledError()
        await many.on_stop()
        agent1.stop.assert_called_once_with()
        agent2.stop.assert_called_once_with()

    def test_hashable(self, *, agents):
        assert isinstance(agents, Hashable)
        assert hash(agents)

    @pytest.mark.asyncio
    async def test_start(self, *, many):
        await many.start()
        for agent in many.values():
            agent.maybe_start.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_restart(self, *, many):
        await many.restart()
        for agent in many.values():
            agent.stop.assert_called_once_with()

    def test_service_reset(self, *, many):
        many.service_reset()
        for agent in many.values():
            agent.service_reset.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_stop(self, *, many):
        await many.stop()
        for agent in many.values():
            agent.cancel.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_wait_until_agents_started(self, *, agents):
        agents.wait_for_stopped = AsyncMock()
        await agents.wait_until_agents_started()
        agents.wait_for_stopped.assert_called_once_with(
            agents._agents_started,
        )

    @pytest.mark.asyncio
    async def test_wait_until_agents_started__producer_only(
            self, *, app, agents):
        app.producer_only = True
        agents._agents_started.clear()
        await agents.wait_until_agents_started()

    @pytest.mark.asyncio
    async def test_wait_until_agents_started__client_only(
            self, *, app, agents):
        app.client_only = True
        agents._agents_started.clear()
        await agents.wait_until_agents_started()

    def test_update_topic_index(self, *, many, agent1, agent2):
        many.update_topic_index()
        assert set(many._by_topic['t1']) == {agent1, agent2}
        assert set(many._by_topic['t2']) == {agent2}
        assert set(many._by_topic['t3']) == {agent2}

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, many, agent1, agent2):
        many.update_topic_index()
        revoked = {TP('t1', 0), TP('t1', 1), TP('t4', 3)}
        newly_assigned = {TP('t2', 0)}
        await many.on_rebalance(
            revoked=revoked,
            newly_assigned=newly_assigned,
        )
        agent1.on_partitions_revoked.assert_called_once_with(
            {TP('t1', 0), TP('t1', 1)},
        )
        agent2.on_partitions_revoked.assert_called_once_with(
            {TP('t1', 0), TP('t1', 1)},
        )
        agent2.on_partitions_assigned.assert_called_once_with(
            newly_assigned,
        )



from unittest.mock import patch

import pytest
from faust import web
from faust.exceptions import ImproperlyConfigured
from faust.sensors.prometheus import PrometheusMonitor
from faust.types import TP
from mode.utils.mocks import Mock, call

TP1 = TP('foo', 3)


class test_PrometheusMonitor:

    def time(self):
        timefun = Mock(name='time()')
        timefun.return_value = 101.1
        return timefun

    @patch('faust.sensors.prometheus.Histogram')
    @patch('faust.sensors.prometheus.Gauge')
    @patch('faust.sensors.prometheus.Counter')
    @patch.object(PrometheusMonitor, 'expose_metrics')
    def prometheus_client(self, app, counter, gauge, histogram, time=None):
        time = time or self.time()

        return PrometheusMonitor(app, time=time)

    @pytest.fixture()
    def stream(self):
        stream = Mock(name='stream')
        stream.shortlabel = 'Stream: Topic: foo'
        return stream

    @pytest.fixture()
    def event(self):
        return Mock(name='event')

    @pytest.fixture()
    def table(self):
        table = Mock(name='table')
        table.name = 'table1'
        return table

    @pytest.fixture()
    def response(self):
        return Mock(name='response', autospec=web.Response)

    @pytest.fixture()
    def view(self):
        return Mock(name='view', autospec=web.View)

    def test_prometheus_client_not_installed(self, app, monkeypatch):
        monkeypatch.setattr('faust.sensors.prometheus.prometheus_client', None)
        with pytest.raises(ImproperlyConfigured):
            PrometheusMonitor(app)

    def test_on_message_in_out(self):
        message = Mock(name='message')
        client = self.prometheus_client()
        client.on_message_in(TP1, 400, message)

        client.messages_received.inc.assert_called_once()
        client.active_messages.inc.assert_called_once()
        client.messages_received_per_topics.labels.assert_called_once_with(
            topic='foo')

        labels = client.messages_received_per_topics_partition.labels
        labels.assert_called_once_with(topic='foo', partition=3)
        labels(topic='foo', partition=3).set.assert_called_once_with(400)

        client.on_message_out(TP1, 400, message)
        client.active_messages.dec.assert_called_once()

    def test_on_stream_event_in_out(self, *, stream, event):
        client = self.prometheus_client()
        state = client.on_stream_event_in(TP1, 401, stream, event)

        client.total_events.inc.assert_called_once()
        client.total_active_events.inc.assert_called_once()
        client.total_events_per_stream.labels.assert_called_once_with(
            stream='stream.topic_foo.events',
        )

        client.on_stream_event_out(TP1, 401, stream, event, state)
        client.total_active_events.dec.assert_called_once()
        client.events_runtime_latency.observe.assert_called_once_with(
            client.secs_to_ms(client.events_runtime[-1]),
        )

    def test_on_table_get(self, table):
        client = self.prometheus_client()
        client.on_table_get(table, 'key')

        client.table_operations.labels.assert_called_once_with(
            table='table.table1', operation='keys_retrieved',
        )
        client.table_operations.labels(
            table='table.table1', operation='keys_retrieved',
        ).inc.assert_called_once()

    def test_on_table_set(self, table):
        client = self.prometheus_client()
        client.on_table_set(table, 'key', 'value')

        client.table_operations.labels.assert_called_once_with(
            table='table.table1', operation='keys_updated',
        )
        client.table_operations.labels(
            table='table.table1', operation='keys_updated',
        ).inc.assert_called_once()

    def test_on_table_del(self, table):
        client = self.prometheus_client()
        client.on_table_del(table, 'key')

        client.table_operations.labels.assert_called_once_with(
            table='table.table1', operation='keys_deleted',
        )
        client.table_operations.labels(
            table='table.table1', operation='keys_deleted',
        ).inc.assert_called_once()

    def test_on_commit_completed(self):
        consumer = Mock(name='consumer')
        client = self.prometheus_client()
        state = client.on_commit_initiated(consumer)
        client.on_commit_completed(consumer, state)

        client.consumer_commit_latency.observe.assert_called_once_with(
            client.ms_since(float(state)))

    def test_on_send_initiated_completed(self):
        producer = Mock(name='producer')
        client = self.prometheus_client()
        state = client.on_send_initiated(
            producer, 'topic1', 'message', 321, 123)

        client.on_send_completed(producer, state, Mock(name='metadata'))

        client.total_sent_messages.inc.assert_called_once()
        client.topic_messages_sent.labels.assert_called_once_with(
            topic='topic.topic1')
        client.topic_messages_sent.labels(
            topic='topic.topic1').inc.assert_called_once()

        client.producer_send_latency.observe.assert_called_once_with(
            client.ms_since(float(state)))

        client.on_send_error(producer, KeyError('foo'), state)

        client.total_error_messages_sent.inc.assert_called()
        client.producer_error_send_latency.observe.assert_called_with(
            client.ms_since(float(state)))

    def test_on_assignment_start_completed(self):
        assignor = Mock(name='assignor')
        client = self.prometheus_client()
        state = client.on_assignment_start(assignor)
        client.on_assignment_completed(assignor, state)

        client.assignment_operations.labels.assert_called_once_with(
            operation=client.COMPLETED)
        client.assignment_operations.labels(
            operation=client.COMPLETED).inc.assert_called_once()
        client.assign_latency.observe.assert_called_once_with(
            client.ms_since(state['time_start']))

    def test_on_assignment_start_failed(self):
        assignor = Mock(name='assignor')
        client = self.prometheus_client()
        state = client.on_assignment_start(assignor)
        client.on_assignment_error(assignor, state, KeyError())

        client.assignment_operations.labels.assert_called_once_with(
            operation=client.ERROR)
        client.assignment_operations.labels(
            operation=client.ERROR).inc.assert_called_once()
        client.assign_latency.observe.assert_called_once_with(
            client.ms_since(state['time_start']))

    def test_on_rebalance(self):
        app = Mock(name='app')
        client = self.prometheus_client()

        state = client.on_rebalance_start(app)
        client.total_rebalances.inc.assert_called_once()

        client.on_rebalance_return(app, state)
        client.total_rebalances.dec.assert_called_once()
        client.total_rebalances_recovering.inc.assert_called()
        client.revalance_done_consumer_latency.observe.assert_called_once_with(
            client.ms_since(state['time_return']))

        client.on_rebalance_end(app, state)
        client.total_rebalances_recovering.dec.assert_called()
        client.revalance_done_latency.observe(
            client.ms_since(state['time_end']))

    def test_on_web_request(self, request, response, view):
        response.status = 404
        self.assert_on_web_request(
            request, response, view, expected_status=404)

    def test_on_web_request_none_response(self, request, view):
        self.assert_on_web_request(request, None, view, expected_status=500)

    def assert_on_web_request(self, request, response, view,
                              expected_status):
        app = Mock(name='app')
        client = self.prometheus_client()
        state = client.on_web_request_start(app, request, view=view)
        client.on_web_request_end(app, request, response, state, view=view)

        client.http_status_codes.labels.assert_called_with(
            status_code=expected_status)
        client.http_status_codes.labels(
            status_code=expected_status).inc.assert_called()
        client.http_latency.observe.assert_called_with(
            client.ms_since(state['time_end']))

    def test_count(self):
        client = self.prometheus_client()
        client.count('metric_name', count=3)

        client.count_metrics_by_name.labels.assert_called_once_with(
            metric='metric_name')
        client.count_metrics_by_name.labels(
            metric='metric_name').inc.assert_called_once_with(3)

    def test_on_tp_commit(self):
        offsets = {
            TP('foo', 0): 1001,
            TP('foo', 1): 2002,
            TP('bar', 3): 3003,
        }
        client = self.prometheus_client()

        client.on_tp_commit(offsets)
        client.topic_partition_offset_commited.labels.assert_has_calls([
            call(topic='foo', partition=0),
            call().set(1001),
            call(topic='foo', partition=1),
            call().set(2002),
            call(topic='bar', partition=3),
            call().set(3003),
        ])

    def test_track_tp_end_offsets(self):
        client = self.prometheus_client()
        client.track_tp_end_offset(TP('foo', 0), 4004)

        client.topic_partition_end_offset.labels.assert_called_once_with(
            topic='foo', partition=0)
        client.topic_partition_end_offset.labels(
            topic='foo', partition=0).set.assert_called_once_with(4004)



# -*- coding: utf-8 -*-
import sys
from contextlib import suppress
from sphinx_celery import conf

extensions = []  # set by build_config
sys.path.append('.')

globals().update(conf.build_config(
    'faust', __file__,
    project='Faust',
    version_dev='1.1',
    version_stable='1.0',
    canonical_url='http://faust.readthedocs.io',
    webdomain='',
    github_project='robinhood/faust',
    copyright='2017-2020',
    html_logo='images/logo.png',
    html_favicon='images/favicon.ico',
    html_prepend_sidebars=[],
    include_intersphinx={'sphinx'},
    extra_extensions=[
        'sphinx.ext.napoleon',
        'alabaster',
        'typehints',
        'faustdocs',
    ],
    extra_intersphinx_mapping={
        'aiohttp': ('https://aiohttp.readthedocs.io/en/stable/', None),
        'aiokafka': ('https://aiokafka.readthedocs.io/en/stable/', None),
        'aredis': ('https://aredis.readthedocs.io/en/latest/', None),
        'click': ('https://click.palletsprojects.com/en/7.x/', None),
        'kafka-python': (
            'https://kafka-python.readthedocs.io/en/master/', None),
        'mode': ('https://mode.readthedocs.io/en/latest/', None),
        'mypy': ('https://mypy.readthedocs.io/en/latest/', None),
        'pytest': ('https://pytest.readthedocs.io/en/latest/', None),
        'python': ('https://docs.python.org/dev/', None),
        'rocksdb': ('https://python-rocksdb.readthedocs.io/en/latest/', None),
        'statsd': ('https://statsd.readthedocs.io/en/latest/', None),
        'terminaltables': ('https://robpol86.github.io/terminaltables/', None),
        'uvloop': ('https://uvloop.readthedocs.io', None),
        'venusian': ('https://venusian.readthedocs.io/en/latest/', None),
        'yarl': ('https://yarl.readthedocs.io/en/latest/', None),
    },
    # django_settings='testproj.settings',
    # from pathlib import Path
    # path_additions=[Path.cwd().parent / 'testproj']
    apicheck_ignore_modules=[
        'faust.__main__',
        'faust.app._attached',
        'faust.assignor',
        'faust.cli',
        'faust.models',
        'faust.serializers',
        'faust.transport.drivers.confluent',
        'faust.types',
        'faust.types._env',
        'faust.utils',
        'faust.utils._iso8601_python',
        r'faust.utils.kafka.*',
        'faust.web',
        r'faust.web.apps.*',
        'faust.web.apps.stats.app',
        'faust.web.apps.router.app',
        'faust'
        'faust.web.drivers',
        r'.*\._cython.*',
    ],
))


def configcheck_project_settings():
    from faust import Settings
    return Settings.setting_names()


html_theme = 'alabaster'
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
    ],
}
html_theme_options = {
    'description': 'A library for building streaming applications in Python.',
    'github_banner': True,
    'travis_button': True,
    'show_related': True,
    'github_user': 'robinhood',
    'pre_bg': '#4c4c4c',
    'github_repo': 'faust',
    'github_type': 'star',
}
templates_path = ['_templates']

autodoc_member_order = 'bysource'

pygments_style = 'monokai'

# This option is deprecated and raises an error.
with suppress(NameError):
    del(html_use_smartypants)  # noqa

napoleon_use_keyword = True

applehelp_bundle_id = 'Faust'
epub_identifier = 'Faust'
latex_engine = 'xelatex'
latex_elements = {
    'inputenc': '',
    'utf8extra': '',
    'sphinxsetup':
        r'verbatimwithframe=false, VerbatimColor={rgb}{0.47, 0.41, 0.47}',
}

ignored_settings = {'ssl_context'}


def configcheck_should_ignore(setting):
    return setting in ignored_settings



"""Web endpoint showing graph of running :pypi:`mode` services."""
import io
from faust import web

__all__ = ['Graph', 'blueprint']

blueprint = web.Blueprint('graph')


@blueprint.route('/', name='detail')
class Graph(web.View):
    """Render image from graph of running services."""

    async def get(self, request: web.Request) -> web.Response:
        """Draw image of the services running in this worker."""
        try:
            import pydot
        except ImportError:
            return self.text('Please install pydot: pip install pydot',
                             status=500)
        o = io.StringIO()
        beacon = self.app.beacon.root or self.app.beacon
        beacon.as_graph().to_dot(o)
        graph, = pydot.graph_from_dot_data(o.getvalue())
        return self.bytes(graph.create_png(), content_type='image/png')



import asyncio
import faust
from faust.types import StreamT, TP
from mode import label
from mode.utils.aiter import aiter, anext
from mode.utils.mocks import Mock
from mode.utils.queues import FlowControlQueue
import pytest
from .helpers import channel_empty, message, times_out


class Point(faust.Record):
    x: int
    y: int


@pytest.fixture
def channel(*, app):
    return app.channel()


def test_repr(*, channel):
    assert repr(channel)


def test_schema__default(*, channel):
    assert channel.key_type is None
    assert channel.value_type is None

    assert channel.schema is not None
    assert channel.schema.key_type is None
    assert channel.schema.value_type is None
    assert channel.schema.key_serializer is None
    assert channel.schema.value_serializer is None


def test_schema__from_schema(*, app):
    schema = faust.Schema(
        key_type='bytes',
        value_type='bytes',
        key_serializer='msgpack',
        value_serializer='msgpack',
    )
    channel = app.channel(schema=schema)

    assert channel.key_type == 'bytes'
    assert channel.value_type == 'bytes'

    assert channel.schema is schema
    assert channel.schema.key_type == channel.key_type
    assert channel.schema.value_type == channel.value_type
    assert channel.schema.key_serializer == 'msgpack'
    assert channel.schema.value_serializer == 'msgpack'


def test_schema__overriding(*, app):
    schema = faust.Schema(
        key_type='bytes',
        value_type='bytes',
        key_serializer='msgpack',
        value_serializer='msgpack',
    )
    channel = app.channel(schema=schema, key_type='str', value_type='str')

    assert channel.key_type == 'str'
    assert channel.value_type == 'str'
    assert channel.schema.key_type == 'str'
    assert channel.schema.value_type == 'str'
    assert channel.schema.key_serializer == 'msgpack'
    assert channel.schema.value_serializer == 'msgpack'


def test_repr__active_partitions_empty(*, channel):
    channel.active_partitions = set()
    assert repr(channel)


def test_repr__with_active_partitions(*, channel):
    channel.active_partitions = {TP('foo', 0), TP('foo', 1)}
    assert repr(channel)


def test_label(*, channel):
    assert label(channel)


def test_str(*, channel):
    assert str(channel)


def test_send_soon(*, channel):
    with pytest.raises(NotImplementedError):
        channel.send_soon(key=b'k', value=b'val')


def test_stream(*, channel):
    s = channel.stream()
    assert isinstance(s, StreamT)
    assert s.channel.queue is not channel.queue
    assert s.channel._root is channel
    assert s.channel in channel._subscribers


def test_get_topic_name(*, channel):
    with pytest.raises(NotImplementedError):
        channel.get_topic_name()


def test_clone(*, app):
    c = app.channel(key_type=Point, value_type=Point, maxsize=99, loop=33)
    assert isinstance(c.queue, FlowControlQueue)
    assert c.key_type is Point
    assert c.value_type is Point
    assert c.maxsize == 99
    assert c.loop == 33
    assert not c.is_iterator

    c2 = c.clone()
    assert c2.key_type is Point
    assert c2.value_type is Point
    assert c2.maxsize == 99
    assert c2.loop == 33
    assert not c2.is_iterator
    assert c2.queue is not c.queue
    assert c2._root is c
    assert c2 not in c._subscribers
    assert c.subscriber_count == 0

    c3 = c2.clone(is_iterator=True)
    assert c3.key_type is Point
    assert c3.value_type is Point
    assert c3.maxsize == 99
    assert c3.loop == 33
    assert c3.is_iterator
    assert c3.queue is not c.queue
    assert c3._root is c
    assert c2._root is c
    assert c3 in c._subscribers
    assert c2 not in c._subscribers
    assert c.subscriber_count == 1


@pytest.mark.asyncio
async def test_send_receive(*, app):
    app.flow_control.resume()
    channel1 = app.channel(maxsize=10)
    channel2 = app.channel(maxsize=1)
    with pytest.raises(RuntimeError):
        # must call aiter
        await channel1.__anext__()
    it1 = aiter(channel1)
    it2 = aiter(channel2)
    await channel2.put(b'xuzzy')
    assert await times_out(channel2.put(b'bar'))  # maxsize=1
    assert it1.queue is not channel1.queue
    assert it1._root is channel1
    assert it1 in channel1._subscribers
    assert channel1.subscriber_count == 1
    assert channel1.queue is not channel2.queue
    assert await channel_empty(channel1)
    await channel1.put(b'foo')
    assert await anext(it1) == b'foo'
    assert await channel_empty(channel1)
    for i in range(10):
        await channel1.put(i)
    assert await times_out(channel1.put(10))  # maxsize=10
    for i in range(10):
        assert await anext(it1) == i
    assert await channel_empty(channel1)
    assert await anext(it2) == b'xuzzy'
    assert await channel_empty(channel2)

    it1_2 = aiter(channel1)
    assert it1_2._root is channel1
    assert it1_2 in channel1._subscribers
    assert channel1.subscriber_count == 2

    await channel1.put(b'moo')

    assert await anext(it1) == b'moo'
    assert await anext(it1_2) == b'moo'


@pytest.mark.asyncio
async def test_on_key_decode_error(*, app):
    channel = app.channel()
    await channel.on_key_decode_error(KeyError('foo'), 'msg')
    with pytest.raises(KeyError):
        await channel.get()


@pytest.mark.asyncio
async def test_on_value_decode_error(*, app):
    channel = app.channel()
    await channel.on_value_decode_error(KeyError('foo'), 'msg')
    with pytest.raises(KeyError):
        await channel.get()


def test_derive(*, app):
    channel = app.channel(maxsize=1)
    assert channel.derive() is channel


@pytest.mark.asyncio
async def test_declare__does_nothing(*, app):
    await app.channel().declare()


def test_clone_using_queue(*, channel):
    queue = asyncio.Queue()
    chan2 = channel.clone_using_queue(queue)
    assert chan2.queue is queue
    assert chan2.is_iterator


@pytest.mark.asyncio
async def test_interface_maybe_declare(*, channel):
    await channel.maybe_declare()


@pytest.mark.asyncio
async def test_decode(*, channel):
    msg = message(b'key', b'value')
    event = await channel.decode(msg)
    assert event.message is msg


@pytest.mark.asyncio
async def test_deliver(*, channel, app):
    app.flow_control.resume()
    msg = message(b'key', b'value')
    queue = channel.queue
    await channel.deliver(msg)
    event = queue.get_nowait()
    assert event.message is msg


def test_as_future_message__eager_partitioning(*, app):
    topic = app.topic('foo')
    app.producer = Mock(name='producer')
    fut = topic.as_future_message(
        key=b'foo',
        value=b'bar',
        partition=None,
        eager_partitioning=True,
    )
    assert fut.message.partition is not None


def test_as_future_message__eager_partitioning_on_channel(*, channel, app):
    app.producer = Mock(name='producer')
    with pytest.raises(NotImplementedError):
        channel.as_future_message(
            key=b'foo',
            value=b'bar',
            partition=None,
            eager_partitioning=True,
        )



"""Producer.

The Producer is responsible for:

   - Holds reference to the transport that created it
   - ... and the app via ``self.transport.app``.
   - Sending messages.
"""
import asyncio
from asyncio import QueueEmpty
from typing import Any, Awaitable, Mapping, Optional, cast
from mode import Seconds, Service
from faust.types import AppT, HeadersArg
from faust.types.tuples import FutureMessage, RecordMetadata, TP
from faust.types.transports import ProducerBufferT, ProducerT, TransportT

__all__ = ['Producer']


class ProducerBuffer(Service, ProducerBufferT):

    max_messages = 100

    def __post_init__(self) -> None:
        self.pending = asyncio.Queue()

    def put(self, fut: FutureMessage) -> None:
        """Add message to buffer.

        The message will be eventually produced, you can await
        the future to wait for that to happen.
        """
        self.pending.put_nowait(fut)

    async def on_stop(self) -> None:
        await self.flush()

    async def flush(self) -> None:
        """Flush all messages (draining the buffer)."""
        get_pending = self.pending.get_nowait
        send_pending = self._send_pending

        if self.size:
            while True:
                try:
                    msg = get_pending()
                except QueueEmpty:
                    break
                else:
                    await send_pending(msg)

    async def flush_atmost(self, n: int) -> int:
        """Flush at most ``n`` messages."""
        get_pending = self.pending.get_nowait
        send_pending = self._send_pending

        if self.size:
            for i in range(n):
                try:
                    msg = get_pending()
                except QueueEmpty:
                    return i
                else:
                    await send_pending(msg)
            return n
        else:
            return 0

    async def _send_pending(self, fut: FutureMessage) -> None:
        await fut.message.channel.publish_message(fut, wait=False)

    async def wait_until_ebb(self) -> None:
        """Wait until buffer is of an acceptable size.

        Modifying a table key is using the Python dictionary API,
        and as ``__getitem__`` is synchronous we have to add
        pending messages to a buffer.

        The ``__getitem__`` method cannot drain the buffer as doing
        so requires trampolining into the event loop.

        To solve this, we have the conductor wait until the buffer
        is of an acceptable size before resuming stream processing flow.
        """
        if self.size > self.max_messages:
            await self.flush_atmost(self.max_messages)

    @Service.task
    async def _handle_pending(self) -> None:
        get_pending = self.pending.get
        send_pending = self._send_pending
        while not self.should_stop:
            msg = await get_pending()
            await send_pending(msg)

    @property
    def size(self) -> int:
        """Current buffer size (messages waiting to be produced)."""
        queue_items = self.pending._queue  # type: ignore
        queue_items = cast(list, queue_items)
        return len(queue_items)


class Producer(Service, ProducerT):
    """Base Producer."""

    app: AppT

    _api_version: str

    def __init__(self, transport: TransportT,
                 loop: asyncio.AbstractEventLoop = None,
                 **kwargs: Any) -> None:
        self.transport = transport
        self.app = self.transport.app
        conf = self.transport.app.conf
        self.client_id = conf.broker_client_id
        self.linger_ms = conf.producer_linger_ms
        self.max_batch_size = conf.producer_max_batch_size
        self.acks = conf.producer_acks
        self.max_request_size = conf.producer_max_request_size
        self.compression_type = conf.producer_compression_type
        self.request_timeout = conf.producer_request_timeout
        self.ssl_context = conf.ssl_context
        self.credentials = conf.broker_credentials
        self.partitioner = conf.producer_partitioner
        api_version = self._api_version = conf.producer_api_version
        assert api_version is not None
        super().__init__(loop=loop or self.transport.loop, **kwargs)

        self.buffer = ProducerBuffer(loop=self.loop, beacon=self.beacon)

    async def on_start(self) -> None:
        await self.add_runtime_dependency(self.buffer)

    async def send(self, topic: str, key: Optional[bytes],
                   value: Optional[bytes],
                   partition: Optional[int],
                   timestamp: Optional[float],
                   headers: Optional[HeadersArg],
                   *,
                   transactional_id: str = None) -> Awaitable[RecordMetadata]:
        """Schedule message to be sent by producer."""
        raise NotImplementedError()

    def send_soon(self, fut: FutureMessage) -> None:
        self.buffer.put(fut)

    async def send_and_wait(self, topic: str, key: Optional[bytes],
                            value: Optional[bytes],
                            partition: Optional[int],
                            timestamp: Optional[float],
                            headers: Optional[HeadersArg],
                            *,
                            transactional_id: str = None) -> RecordMetadata:
        """Send message and wait for it to be transmitted."""
        raise NotImplementedError()

    async def flush(self) -> None:
        """Flush all in-flight messages."""
        # XXX subclasses must call self.buffer.flush() here.
        ...

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 1000.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        raise NotImplementedError()

    def key_partition(self, topic: str, key: bytes) -> TP:
        """Hash key to determine partition."""
        raise NotImplementedError()

    async def begin_transaction(self, transactional_id: str) -> None:
        """Begin transaction by id."""
        raise NotImplementedError()

    async def commit_transaction(self, transactional_id: str) -> None:
        """Commit transaction by id."""
        raise NotImplementedError()

    async def abort_transaction(self, transactional_id: str) -> None:
        """Abort and rollback transaction by id."""
        raise NotImplementedError()

    async def stop_transaction(self, transactional_id: str) -> None:
        """Stop transaction by id."""
        raise NotImplementedError()

    async def maybe_begin_transaction(self, transactional_id: str) -> None:
        """Begin transaction by id, if not already started."""
        raise NotImplementedError()

    async def commit_transactions(
            self,
            tid_to_offset_map: Mapping[str, Mapping[TP, int]],
            group_id: str,
            start_new_transaction: bool = True) -> None:
        """Commit transactions."""
        raise NotImplementedError()

    def supports_headers(self) -> bool:
        """Return :const:`True` if headers are supported by this transport."""
        return False



from decimal import Decimal
import faust
from faust.types import StreamT
from faustapp.app import app
from .models import Account


class AccountRecord(faust.Record):
    name: str
    score: float
    active: bool


@app.agent()
async def add_account(accounts: StreamT[AccountRecord]):
    async for account in accounts:
        result = Account.objects.create(
            name=account.name,
            score=Decimal(str(account.score)),
            active=account.active,
        )
        yield result.pk


@app.agent()
async def disable_account(account_ids: StreamT[int]):
    async for account_id in account_ids:
        account = Account.objects.get(pk=account_id)
        account.active = False
        account.save()
        yield account.active



import pytest
from faust.exceptions import SameNode
from mode.utils.mocks import Mock


@pytest.yield_fixture()
def web_client(loop, aiohttp_client, web):
    try:
        yield aiohttp_client(web.web_app)
    finally:
        # Cleanup threads started by loop.run_in_executor
        # at shutdown.
        if loop._default_executor is not None:
            loop._default_executor.shutdown()


@pytest.fixture()
def router_same(app):
    app.router.route_req = Mock(name='app.router.route_req')
    app.router.route_req.side_effect = SameNode()
    return app.router



from datetime import datetime, timedelta
from time import time
import random
import faust


class RawModel(faust.Record):
    date: datetime
    value: float


class AggModel(faust.Record):
    date: datetime
    count: int
    mean: float


TOPIC = 'raw-event'
SINK = 'agg-event'
TABLE = 'tumbling_table'
KAFKA = 'kafka://localhost:9092'
CLEANUP_INTERVAL = 1.0
WINDOW = 10
WINDOW_EXPIRES = 1
PARTITIONS = 1

app = faust.App('windowed-agg', broker=KAFKA, version=1, topic_partitions=1)

app.conf.table_cleanup_interval = CLEANUP_INTERVAL
source = app.topic(TOPIC, value_type=RawModel)
sink = app.topic(SINK, value_type=AggModel)


def window_processor(key, events):
    timestamp = key[1][0]
    values = [event.value for event in events]
    count = len(values)
    mean = sum(values) / count

    print(
        f'processing window:'
        f'{len(values)} events,'
        f'mean: {mean:.2f},'
        f'timestamp {timestamp}',
    )

    sink.send_soon(value=AggModel(date=timestamp, count=count, mean=mean))


tumbling_table = (
    app.Table(
        TABLE,
        default=list,
        partitions=PARTITIONS,
        on_window_close=window_processor,
    )
    .tumbling(WINDOW, expires=timedelta(seconds=WINDOW_EXPIRES))
    .relative_to_field(RawModel.date)
)


@app.agent(source)
async def print_windowed_events(stream):
    async for event in stream:
        value_list = tumbling_table['events'].value()
        value_list.append(event)
        tumbling_table['events'] = value_list


@app.timer(0.1)
async def produce():
    await source.send(value=RawModel(value=random.random(), date=int(time())))


if __name__ == '__main__':
    app.main()



"""Actor - Individual Agent instances."""
from typing import Any, AsyncGenerator, AsyncIterator, Coroutine, Set, cast
from mode import Service
from mode.utils.tracebacks import format_agen_stack, format_coro_stack

from faust.types import ChannelT, StreamT, TP
from faust.types.agents import (
    ActorT,
    AgentT,
    AsyncIterableActorT,
    AwaitableActorT,
    _T,
)
__all__ = ['Actor', 'AsyncIterableActor', 'AwaitableActor']


class Actor(ActorT, Service):
    """An actor is a specific agent instance."""

    mundane_level = 'debug'

    # Agent will start n * concurrency actors.

    def __init__(self,
                 agent: AgentT,
                 stream: StreamT,
                 it: _T,
                 index: int = None,
                 active_partitions: Set[TP] = None,
                 **kwargs: Any) -> None:
        self.agent = agent
        self.stream = stream
        self.it = it
        self.index = index
        self.active_partitions = active_partitions
        self.actor_task = None
        Service.__init__(self, **kwargs)

    async def on_start(self) -> None:
        """Call when actor is starting."""
        assert self.actor_task
        self.add_future(self.actor_task)

    async def on_stop(self) -> None:
        """Call when actor is being stopped."""
        self.cancel()

    async def on_isolated_partition_revoked(self, tp: TP) -> None:
        """Call when an isolated partition is being revoked."""
        self.log.debug('Cancelling current task in actor for partition %r', tp)
        self.cancel()
        self.log.info('Stopping actor for revoked partition %r...', tp)
        await self.stop()
        self.log.debug('Actor for revoked partition %r stopped')

    async def on_isolated_partition_assigned(self, tp: TP) -> None:
        """Call when an isolated partition is being assigned."""
        self.log.dev('Actor was assigned to %r', tp)

    def cancel(self) -> None:
        """Tell actor to stop reading from the stream."""
        cast(ChannelT, self.stream.channel)._throw(StopAsyncIteration())

    def __repr__(self) -> str:
        return f'<{self.shortlabel}>'

    @property
    def label(self) -> str:
        """Return human readable description of actor."""
        s = self.agent._agent_label(name_suffix='*')
        if self.stream.active_partitions:
            partitions = {
                tp.partition for tp in self.stream.active_partitions
            }
            s += f' isolated={partitions}'
        return s


class AsyncIterableActor(AsyncIterableActorT, Actor):
    """Used for agent function that yields."""

    def __aiter__(self) -> AsyncIterator:
        return self.it.__aiter__()

    def traceback(self) -> str:
        return format_agen_stack(cast(AsyncGenerator, self.it))


class AwaitableActor(AwaitableActorT, Actor):
    """Used for actor function that do not yield."""

    def __await__(self) -> Any:
        return self.it.__await__()

    def traceback(self) -> str:
        return format_coro_stack(cast(Coroutine, self.it))



from collections import Counter, defaultdict
from . import states
from .app import get_error_topic, get_reporting_app, get_reporting_topic
from .web import env

app = get_reporting_app()

error_topic = get_error_topic(app)
reporting_topic = get_reporting_topic(app)

report_index = {}
report_by_category = defaultdict(lambda: defaultdict(dict))
unique_servers = set()
state_counts = Counter()

errors = defaultdict(lambda: defaultdict(list))

color_for_state = {
    states.OK: '#007BFF',
    states.FAIL: '#19A9D5',
    states.SLOW: '#8fc9fb',
    states.STALL: '#1909E5',
    states.UNASSIGNED: '#003f3c',
    states.REBALANCING: '#005f3c',
    states.PAUSED: '#003f3c',
}


@app.agent(reporting_topic)
async def process_report(reports):
    async for r in reports:
        report_by_category[r.app_id][r.category][r.hostname] = r.details
        previous = report_index.get(r.key)
        report_index[r.key] = r
        if previous is not None and previous.state != r.state:
            print(f'{r.key} changed {previous.state} -> {r.state}')
        unique_servers.add((r.app_id, r.hostname))
        state_counts[r.state] += 1


@app.agent(error_topic)
async def process_error(stream):
    async for error in stream:
        errors[error.app_id][error.hostname].append(error)


@app.page('/dashboard/')
async def dashboard(web, request, template_name='dashboard.html'):
    template = env.get_template(template_name)
    return web.html(template.render({
        'total_workers': len(unique_servers),
        'report': report_by_category,
        'errors': errors,
        'state_counts': [
            {
                'name': state,
                'count': count,
                'color': color_for_state[state],
            }
            for state, count in state_counts.items()
        ],
    }))


@app.page('/dashboard/api/')
async def dashboard_json(web, request):
    return web.json({'status': report_by_category})


if __name__ == '__main__':
    app.main()



"""Locals - Current test & execution context."""
import typing
from typing import Optional
from mode.locals import LocalStack
from .models import TestExecution

if typing.TYPE_CHECKING:
    from .runners import TestRunner as _TestRunner
else:  # pragma: no cover
    class _TestRunner: ...   # noqa

__all__ = [
    'current_execution',
    'current_execution_stack',
    'current_test',
    'current_test_stack',
]

current_test_stack: LocalStack[TestExecution]
current_test_stack = LocalStack()

current_execution_stack: LocalStack[_TestRunner]
current_execution_stack = LocalStack()


def current_execution() -> Optional[_TestRunner]:
    """Return the current :class:`~faust.livecheck.TestRunner`."""
    return current_execution_stack.top


def current_test() -> Optional[TestExecution]:
    """Return information about the current test (if any)."""
    return current_test_stack.top



import os
import threading
import time
from http import HTTPStatus
from typing import Any, NamedTuple
import pytest
from _pytest.assertion.util import _compare_eq_dict, _compare_eq_set
from aiohttp.client import ClientError, ClientSession
from aiohttp.web import Response
from mode.utils.futures import all_tasks
from mode.utils.mocks import (
    AsyncContextManagerMock,
    AsyncMock,
    MagicMock,
    Mock,
    patch,
)

sentinel = object()


class DirtyTest(Exception):
    ...


@pytest.fixture()
def patching(monkeypatch, request):
    """Monkeypath.setattr shortcut.

    Example:
        .. sourcecode:: python

        def test_foo(patching):
            # execv value here will be mock.MagicMock by default.
            execv = patching('os.execv')

            patching('sys.platform', 'darwin')  # set concrete value
            patching.setenv('DJANGO_SETTINGS_MODULE', 'x.settings')

            # val will be of type mock.MagicMock by default
            val = patching.setitem('path.to.dict', 'KEY')
    """
    return _patching(monkeypatch, request)


@pytest.fixture()
def loop(event_loop):
    return event_loop


class _patching(object):

    def __init__(self, monkeypatch, request):
        self.monkeypatch = monkeypatch
        self.request = request

    def __getattr__(self, name):
        return getattr(self.monkeypatch, name)

    def __call__(self, path, value=sentinel, name=None,
                 new=MagicMock, **kwargs):
        value = self._value_or_mock(value, new, name, path, **kwargs)
        self.monkeypatch.setattr(path, value)
        return value

    def _value_or_mock(self, value, new, name, path, **kwargs):
        if value is sentinel:
            value = new(name=name or path.rpartition('.')[2])
        for k, v in kwargs.items():
            setattr(value, k, v)
        return value

    def setattr(self, target, name=sentinel, value=sentinel, **kwargs):
        # alias to __call__ with the interface of pytest.monkeypatch.setattr
        if value is sentinel:
            value, name = name, None
        return self(target, value, name=name)

    def setitem(self, dic, name, value=sentinel, new=MagicMock, **kwargs):
        # same as pytest.monkeypatch.setattr but default value is MagicMock
        value = self._value_or_mock(value, new, name, dic, **kwargs)
        self.monkeypatch.setitem(dic, name, value)
        return value


class TimeMarks(NamedTuple):
    time: float = None
    monotonic: float = None


@pytest.yield_fixture()
def freeze_time(event_loop, request):
    marks = request.node.get_closest_marker('time')
    timestamp = time.time()
    monotimestamp = time.monotonic()

    with patch('time.time') as time_:
        with patch('time.monotonic') as monotonic_:
            options = TimeMarks(**{
                **{'time': timestamp,
                   'monotonic': monotimestamp},
                **((marks.kwargs or {}) if marks else {}),
            })
            time_.return_value = options.time
            monotonic_.return_value = options.monotonic
            yield options


class SessionMarker(NamedTuple):
    status_code: int
    text: bytes
    json: Any
    json_iterator: Any
    max_failures: int


@pytest.fixture()
def mock_http_client(*, app, monkeypatch, request) -> ClientSession:
    marker = request.node.get_closest_marker('http_session')
    options = SessionMarker(**{
        **{
            'status_code': HTTPStatus.OK,
            'text': b'',
            'json': None,
            'json_iterator': None,
            'max_failures': None,
        },
        **(marker.kwargs or {} if marker else {}),
    })

    def raise_for_status():
        if options.max_failures is not None:
            if session.request.call_count >= options.max_failures:
                return
        if 400 <= options.status_code:
            raise ClientError()

    response = AsyncMock(
        autospec=Response,
        text=AsyncMock(return_value=options.text),
        read=AsyncMock(return_value=options.text),
        json=AsyncMock(
            return_value=options.json,
            side_effect=options.json_iterator,
        ),
        status_code=options.status_code,
        raise_for_status=raise_for_status,
    )
    session = Mock(
        name='http_client',
        autospec=ClientSession,
        request=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        get=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        post=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        put=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        delete=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        patch=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        options=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
        head=Mock(
            return_value=AsyncContextManagerMock(
                return_value=response,
            ),
        ),
    )
    session.marks = options
    monkeypatch.setattr(app, '_http_client', session)
    return session


@pytest.fixture(scope='session', autouse=True)
def _collected_environ():
    return dict(os.environ)


@pytest.yield_fixture(autouse=True)
def _verify_environ(_collected_environ):
    try:
        yield
    finally:
        new_environ = dict(os.environ)
        current_test = new_environ.pop('PYTEST_CURRENT_TEST', None)
        old_environ = dict(_collected_environ)
        old_environ.pop('PYTEST_CURRENT_TEST', None)
        if new_environ != old_environ:
            raise DirtyTest(
                'Left over environment variables',
                current_test,
                _compare_eq_dict(new_environ, old_environ, verbose=2))


def alive_threads():
    return {thread for thread in threading.enumerate() if thread.is_alive()}


@pytest.fixture(scope='session', autouse=True)
def _recorded_threads_at_startup(request):
    try:
        request.session._threads_at_startup
    except AttributeError:
        request.session._threads_at_startup = alive_threads()


@pytest.fixture(autouse=True)
def threads_not_lingering(request):
    try:
        yield
    finally:
        threads_then = request.session._threads_at_startup
        threads_now = alive_threads()
        if threads_then != threads_now:
            request.session._threads_at_startup = threads_now
            raise DirtyTest(
                'Left over threads',
                os.environ.get('PYTEST_CURRENT_TEST'),
                _compare_eq_set(threads_now, threads_then, verbose=2))


@pytest.fixture(autouse=True)
def _recorded_tasks_at_startup(request, loop):
    try:
        request.node._tasks_at_startup
    except AttributeError:
        request.node._tasks_at_startup = set(all_tasks(loop=loop))


@pytest.fixture(autouse=True)
def tasks_not_lingering(request, loop, event_loop, _recorded_tasks_at_startup):
    allow_lingering_tasks = False
    allow_count = 0
    marks = request.node.get_closest_marker('allow_lingering_tasks')
    if marks:
        allow_lingering_tasks = True
        allow_count = marks.kwargs.get('count', 0)
    try:
        yield
    finally:
        tasks_then = request.node._tasks_at_startup
        tasks_now = set(all_tasks(loop=loop))
        if tasks_then != tasks_now:
            request.node._tasks_at_startup = tasks_now
            pending = {task for task in tasks_now if task and not task.done()}
            if pending:
                diff = len(pending - tasks_then)
                if not allow_lingering_tasks or diff > allow_count:
                    raise DirtyTest(
                        'Left over tasks',
                        os.environ.get('PYTEST_CURRENT_TEST'),
                        _compare_eq_set(tasks_now, tasks_then, verbose=2))


def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        'allow_lingering_tasks: Allow test to start background tasks',
    )
    config.addinivalue_line(
        'markers',
        'allow_lingering_tasks: Allow test to start background tasks',
    )
    config.addinivalue_line(
        'markers',
        'time: Set the current time',
    )
    config.addinivalue_line(
        'markers',
        'http_session: Set mock aiohttp session result',
    )



import typing
from decimal import Decimal

import faust
import pytest
from faust.exceptions import KeyDecodeError, ValueDecodeError
from faust.utils import json
from mode.utils.mocks import Mock


class Case(typing.NamedTuple):
    payload: typing.Any
    key_type: typing.Any
    serializer: str
    expected: typing.Any


class User(faust.Record):
    id: str
    first_name: str
    last_name: str


class Account(faust.Record):
    id: str
    active: bool
    user: User


# Account1
USER1 = User('A2', 'George', 'Costanza')
ACCOUNT1 = Account(id='A1', active=True, user=USER1)
ACCOUNT1_JSON = ACCOUNT1.dumps(serializer='json')

# Account1 without blessed key
ACCOUNT1_UNBLESSED = ACCOUNT1.to_representation()
ACCOUNT1_UNBLESSED.pop('__faust')
ACCOUNT1_UNBLESSED_JSON = json.dumps(ACCOUNT1_UNBLESSED)

# Account1 with extra fields
ACCOUNT1_EXTRA_FIELDS = ACCOUNT1.to_representation()
ACCOUNT1_EXTRA_FIELDS.update(
    join_date='12321321312',
    foo={'a': 'A', 'b': 'B'},
    bar=[1, 2, 3, 4],
)
ACCOUNT1_EXTRA_FIELDS_JSON = json.dumps(ACCOUNT1_EXTRA_FIELDS)

# Account2
USER2 = User('B2', 'Elaine', 'Benes')
ACCOUNT2 = Account(id='B1', active=True, user=USER2)
ACCOUNT2_JSON = ACCOUNT2.dumps(serializer='json')

# Json data not made by Faust.
NONFAUST = {
    'str': 'str',
    'int': 1,
    'float': 0.30,
    'lst': [1, 2, 3],
    'dct': {'a': 1, 'b': 2},
}

A_BYTE_STR = b'the quick brown fox'
A_STR_STR = 'the quick brown fox'

VALUE_TESTS = [
    # ### value_type=None, serializer='json'
    #   autodetects blessed records -> Account
    Case(ACCOUNT1_JSON, None, 'json', ACCOUNT1),
    Case(ACCOUNT1_EXTRA_FIELDS_JSON, None, 'json', ACCOUNT1),
    Case(ACCOUNT2_JSON, None, 'json', ACCOUNT2),

    # user field in unblessed model still ends up being user
    # because of the type annotation.
    Case(ACCOUNT1_UNBLESSED_JSON, Account, 'json', ACCOUNT1),

    #   source json dict mapping -> mapping
    Case(json.dumps(NONFAUST), None, 'json', NONFAUST),

    #   source json string payload -> str
    Case(json.dumps(A_STR_STR), None, 'json', A_STR_STR),

    # ### value_type=None, serializer='raw'
    #   source unicode payload -> bytes
    Case(A_STR_STR, None, 'raw', A_BYTE_STR),

    #   source bytes payload -> bytes
    Case(A_BYTE_STR, None, 'raw', A_BYTE_STR),

    # ### value_type=bytes, serializer='json'
    #   source json str -> bytes
    Case(json.dumps(A_STR_STR), bytes, 'json', A_BYTE_STR),

    # ### value_type=str, serializer='json'
    #   source json str -> str
    Case(json.dumps(A_STR_STR), str, 'json', A_STR_STR),

    # ###  value_type=bytes, serializer='raw'
    #  source bytes -> bytes
    Case(A_BYTE_STR, bytes, 'raw', A_BYTE_STR),

    # ### value_type=str, serializer='raw'
    #   source str gives str
    Case(A_STR_STR, str, 'raw', A_STR_STR),

    # ### value_type=Account, serializer='json'
    #   source Account json -> Account
    Case(ACCOUNT1_JSON, Account, 'json', ACCOUNT1),
    Case(ACCOUNT1_EXTRA_FIELDS_JSON, Account, 'json', ACCOUNT1),

    # source non blessed json -> Account
    Case(ACCOUNT1_UNBLESSED_JSON, Account, 'json', ACCOUNT1),

    # source blessed User json -> User (!!!)
    # value_type is None so it accepts any type
    Case(USER1.dumps(serializer='json'), None, 'json', USER1),

    # key/value=None
    Case(None, None, 'json', None),
]


@pytest.mark.parametrize('payload,typ,serializer,expected', VALUE_TESTS)
def test_loads_key(payload, typ, serializer, expected, *, app):
    assert app.serializers.loads_key(
        typ, payload, serializer=serializer) == expected


def test_loads_key__expected_model_received_None(*, app):
    with pytest.raises(KeyDecodeError):
        app.serializers.loads_key(Account, None, serializer='json')


def test_loads_key__propagates_MemoryError(*, app):
    app.serializers._loads = Mock(name='_loads')
    app.serializers._loads.side_effect = MemoryError()
    with pytest.raises(MemoryError):
        app.serializers.loads_key(Account, ACCOUNT1_JSON, serializer='json')


def test_loads_value__propagates_MemoryError(*, app):
    app.serializers._loads = Mock(name='_loads')
    app.serializers._loads.side_effect = MemoryError()
    with pytest.raises(MemoryError):
        app.serializers.loads_value(Account, ACCOUNT1_JSON, serializer='json')


def test_loads_value__expected_model_received_None(*, app):
    with pytest.raises(ValueDecodeError):
        app.serializers.loads_value(Account, None, serializer='json')


@pytest.mark.parametrize('payload,typ,serializer,expected', VALUE_TESTS)
def test_loads_value(payload, typ, serializer, expected, *, app):
    assert app.serializers.loads_value(
        typ, payload, serializer=serializer) == expected


def test_loads_value_missing_key_raises_error(*, app):
    account = ACCOUNT1.to_representation()
    account.pop('active')
    with pytest.raises(ValueDecodeError):
        app.serializers.loads_value(
            Account, json.dumps(account), serializer='json')


def test_loads_key_missing_key_raises_error(*, app):
    account = ACCOUNT1.to_representation()
    account.pop('active')
    with pytest.raises(KeyDecodeError):
        app.serializers.loads_key(
            Account, json.dumps(account), serializer='json')


def test_dumps_value__bytes(*, app):
    assert app.serializers.dumps_value(
        bytes, b'foo', serializer='json') == b'foo'


@pytest.mark.parametrize('typ,alt,expected', [
    (str, (), 'raw'),
    (bytes, (), 'raw'),
    (str, ('json',), 'json'),
    (bytes, ('json',), 'json'),
    (None, (), None),
])
def test_serializer_type(typ, alt, expected, *, app):
    assert app.serializers._serializer(typ, *alt) == expected


@pytest.mark.parametrize('typ,value,expected_value', [
    (int, '23', 23),
    (float, '23.32', 23.32),
    (Decimal, '23.32', Decimal(23.32)),
    (str, 'foo', 'foo'),
    (bytes, 'foo', b'foo'),
])
def test_prepare_payload(typ, value, expected_value, *, app):
    app.serializers._prepare_payload(typ, value) == expected_value



"""Functional utilities."""
from functools import reduce
from itertools import groupby
from typing import (
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)
from mode.utils.typing import Deque

__all__ = [
    'consecutive_numbers',
    'deque_prune',
    'deque_pushpopmax',
    'translate',
]

T = TypeVar('T')


def consecutive_numbers(it: Iterable[int]) -> Iterator[Sequence[int]]:
    """Find runs of consecutive numbers.

    Notes:
        See https://docs.python.org/2.6/library/itertools.html#examples
    """
    for _, g in groupby(enumerate(it), lambda a: a[0] - a[1]):
        yield [a[1] for a in g]


def deque_prune(l: Deque[T], max: int = None) -> Optional[T]:
    """Prune oldest element in deque if size exceeds ``max``."""
    if max is not None:
        size = len(l)
        if size > max:
            return l.popleft()
    return None


def deque_pushpopmax(l: Deque[T], item: T, max: int = None) -> Optional[T]:
    """Append to deque and remove oldest element if size exceeds ``max``."""
    l.append(item)
    return deque_prune(l, max)


def translate(table: Mapping, s: str) -> str:
    """Replace characters and patterns in string ``s``.

    Works similar to :meth:`str.translate`, but replacements and patterns
    can be full length strings instead of character by character.

    Arguments:
        table: A mapping of characters/patterns to their replacement string.
        s: The string to translate

    Note:
        Table is the first argument in the signature for compatibility
        with :func:`~functools.partial`:

        .. sourcecode:: pycon

           >>> t = partial(translate, {'.': '_'})
           >>> t('foo.bar')
           'foo_bar'

    Examples:
        >>> translate('foo.bar@baz', {'.': '_', '@': '.'})
        'foo_bar.baz'
    """

    def on_reduce(acc: str, kv: Tuple[str, str]) -> str:
        return acc.replace(kv[0], kv[1])  # table key/value

    return reduce(on_reduce, table.items(), s)
